import logging
from unittest import mock

from bx_py_utils.test_utils.datetime import parse_dt
from django.contrib.messages import get_messages
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from bx_django_utils.test_utils.datetime import MockDatetimeGenerator
from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin
from bx_django_utils.test_utils.users import make_minimal_test_user
from bx_django_utils_tests.approve_workflow_test_app.models import ApproveTestModel, RelatedApproveTestModel


class BaseApproveModelTestCase(TestCase):
    @mock.patch.object(timezone, 'now', MockDatetimeGenerator())
    def test_approve(self):
        draft = ApproveTestModel.objects.create(
            id='00000000-0000-0000-0000-000000000001',
            name='v1',
            title='',
        )
        assert draft.is_draft is True
        assert draft.approved is None
        assert draft.is_approved() is False
        assert draft.create_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert draft.update_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert draft.blocked is False
        assert draft.ready_to_approve is False
        assert draft.get_missing_field_info() == [('title', 'A Title', '')]

        # We can't approve with required fields:
        with self.assertLogs('bx_django_utils', level=logging.WARNING) as logs:
            with self.assertRaisesMessage(ValidationError, "{'title': ['This field cannot be blank.']}"):
                draft.approve()

        # Not set "ready_to_approve" generates only a warning:
        assert logs.output == [
            "WARNING:bx_django_utils.approve_workflow.models:Not ready to approve:"
            " pk:UUID('00000000-0000-0000-0000-000000000001') model:ApproveTestModel"
        ]

        # Set a 'title' and check missing fields:
        draft.title = 'v1'
        draft.ready_to_approve = True
        draft.save()
        draft = ApproveTestModel.objects.get(pk=draft.pk)
        assert draft.create_dt == parse_dt('2001-01-01T00:00:00+0000')  # is the same?
        assert draft.update_dt == parse_dt('2002-01-01T00:00:00+0000')  # updated?
        assert draft.get_missing_field_info() == []

        # "blocked" entries can't approve:

        draft.blocked = True
        with self.assertLogs('bx_django_utils', level=logging.DEBUG) as logs:
            with self.assertRaisesMessage(ValidationError, 'Blocked entries can not approved!'):
                draft.approve()
        assert ApproveTestModel.objects.count() == 1  # only one draft exists
        assert logs.output == [
            'DEBUG:bx_django_utils.approve_workflow.models:Approve <ApproveTestModel '
            'pk:00000000-0000-0000-0000-000000000001 (draft, not ready)>',
        ]

        # ready & not blocked -> approve will work:
        draft.blocked = False
        with self.assertLogs('bx_django_utils', level=logging.WARNING) as logs:
            approved = draft.approve()
        assert isinstance(approved, ApproveTestModel)
        assert logs.output == [
            "WARNING:bx_django_utils.approve_workflow.models:Not ready to approve:"
            " pk:UUID('00000000-0000-0000-0000-000000000001') model:ApproveTestModel"
        ]

        # Now we have two entries: The draft and the approve one:
        assert tuple(
            ApproveTestModel.objects.values_list('pk', 'title', 'is_draft').order_by('-is_draft')
        ) == (
            (draft.pk, 'v1', True),
            (approved.pk, 'v1', False)
        )

        # Get fresh versions from database and check:

        approved = ApproveTestModel.objects.get(pk=approved.pk)
        draft = ApproveTestModel.objects.get(pk=draft.pk)

        assert draft.is_draft
        assert not approved.is_draft
        assert draft.approved.pk == approved.pk

        # Both entries are "is_approved" (Used in admin for drafts, too):
        assert draft.is_approved() is True
        assert approved.is_approved() is True

        # Note: After approve the "ready" flag will be removed (for next edit):
        assert draft.ready_to_approve is False
        assert approved.ready_to_approve is False

        assert approved.title == draft.title == 'v1'

        # The Approved version should have the same timestamps:
        assert approved.create_dt == draft.create_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert approved.update_dt == draft.update_dt == parse_dt('2002-01-01T00:00:00+0000')

        # Only drafts can be approved:

        with self.assertRaisesMessage(ValidationError, 'Only drafts can be approve!'):
            approved.approve()

        # Change a field:

        draft.title = 'v2'
        draft.ready_to_approve = True
        draft.save()

        assert tuple(
            ApproveTestModel.objects.values_list('pk', 'title', 'is_draft').order_by('-is_draft')
        ) == (
            (draft.pk, 'v2', True),
            (approved.pk, 'v1', False)
        )

        # Publish again, with ready_to_approve = True -> No warning

        with self.assertLogs('bx_django_utils', level=logging.INFO) as logs:
            approved = draft.approve()
        assert logs.output == [
            'INFO:bx_django_utils.approve_workflow.models:<ApproveTestModel '
            'pk:00000000-0000-0000-0000-000000000001 (draft, not ready)> was approved to '
            f'<ApproveTestModel pk:{approved.pk} (approved)>',
        ]

        assert tuple(
            ApproveTestModel.objects.values_list('pk', 'title', 'is_draft').order_by('-is_draft')
        ) == (
            (draft.pk, 'v2', True),
            (approved.pk, 'v2', False)
        )

        # Get fresh versions from database and check:

        approved = ApproveTestModel.objects.get(pk=approved.pk)
        draft = ApproveTestModel.objects.get(pk=draft.pk)

        assert draft.is_draft
        assert draft.approved.pk == approved.pk
        assert not approved.is_draft

        # Both entries are "is_approved" (Used in admin for drafts, too):
        assert draft.is_approved() is True
        assert approved.is_approved() is True

        # Note: After approve the "ready" flag will be removed (for next edit):
        assert draft.ready_to_approve is False
        assert approved.ready_to_approve is False

        assert approved.title == draft.title == 'v2'

        # Add a relation:

        relation1_draft = RelatedApproveTestModel.objects.create(
            relation_name='relation one v1',
            relation_title=None,
            main_entry=draft,
        )
        relation2_draft = RelatedApproveTestModel.objects.create(
            relation_name='relation two v1',
            relation_title='The Title',
            main_entry=draft,
        )

        assert tuple(
            RelatedApproveTestModel.objects.values_list(
                'main_entry_id', 'relation_name', 'is_draft'
            ).order_by('-is_draft', 'relation_name')
        ) == (
            (draft.pk, 'relation one v1', True),
            (draft.pk, 'relation two v1', True),
        )

        # Can't publish if relations has missing fields, too:

        with self.assertLogs('bx_django_utils', level=logging.WARNING) as logs:
            with self.assertRaisesMessage(ValidationError, "{'relation_title': ['This field cannot be blank.']}"):
                draft.approve()
        assert logs.output == [
            "WARNING:bx_django_utils.approve_workflow.models:Not ready to approve:"
            " pk:UUID('00000000-0000-0000-0000-000000000001') model:ApproveTestModel"
        ]

        assert relation1_draft.approved is None
        relation1_draft.relation_title = 'Now a title set'
        relation1_draft.save()

        # Publish with relations
        draft.title = 'v3'
        draft.ready_to_approve = True
        draft.save()
        assert tuple(
            ApproveTestModel.objects.values_list('pk', 'title', 'is_draft').order_by('-is_draft')
        ) == (
            (draft.pk, 'v3', True),
            (approved.pk, 'v2', False)
        )
        draft.ready_to_approve = True
        with self.assertLogs('bx_django_utils', level=logging.INFO):
            approved = draft.approve()

        assert tuple(
            ApproveTestModel.objects.values_list('pk', 'title', 'is_draft').order_by('-is_draft')
        ) == (
            (draft.pk, 'v3', True),
            (approved.pk, 'v3', False)
        )
        assert tuple(
            RelatedApproveTestModel.objects.values_list(
                'main_entry_id', 'relation_name', 'relation_title', 'is_draft'
            ).order_by('-is_draft', 'relation_name')
        ) == (
            (draft.pk, 'relation one v1', 'Now a title set', True),
            (draft.pk, 'relation two v1', 'The Title', True),
            (approved.pk, 'relation one v1', 'Now a title set', False),
            (approved.pk, 'relation two v1', 'The Title', False),
        )
        relation1_draft = RelatedApproveTestModel.objects.get(pk=relation1_draft.pk)
        relation1_approved = relation1_draft.approved
        assert relation1_approved is not None
        assert relation1_draft.relation_name == relation1_approved.relation_name == 'relation one v1'

        # Change the draft version doesn't touch the approve one:

        relation1_draft.relation_name = 'relation one v2'
        relation1_draft.save()
        relation1_draft = RelatedApproveTestModel.objects.get(pk=relation1_draft.pk)
        assert relation1_draft.approved.relation_name == 'relation one v1'

        # Remove a relation:
        relation2_draft.delete()

        assert tuple(
            RelatedApproveTestModel.objects.values_list(
                'main_entry_id', 'relation_name', 'is_draft'
            ).order_by('-is_draft', 'relation_name')
        ) == (
            (draft.pk, 'relation one v2', True),
            (approved.pk, 'relation one v1', False),
            (approved.pk, 'relation two v1', False),
        )

        # Publish the new relations:

        draft.ready_to_approve = True
        with self.assertLogs('bx_django_utils', level=logging.INFO):
            draft.approve()

        assert tuple(
            RelatedApproveTestModel.objects.values_list(
                'main_entry_id', 'relation_name', 'is_draft'
            ).order_by('-is_draft', 'relation_name')
        ) == (
            (draft.pk, 'relation one v2', True),
            (approved.pk, 'relation one v2', False),
        )


class BaseApproveModelAdminTestCase(HtmlAssertionMixin, TestCase):
    def test_approve(self):
        draft = ApproveTestModel.objects.create(
            id='00000000-0000-0000-0000-000000000001',
            name='v1',
            title='',
        )
        relation_draft = RelatedApproveTestModel.objects.create(
            id='00000000-0000-0000-0000-000000000002',
            relation_name='relation one v1',
            relation_title=None,
            main_entry=draft,
        )
        minimal_test_user = make_minimal_test_user(
            is_staff=True,
            permissions=(
                'approve_workflow_test_app.add_approvetestmodel',
                'approve_workflow_test_app.approve_approvetestmodel',
                'approve_workflow_test_app.change_approvetestmodel',
                'approve_workflow_test_app.delete_approvetestmodel',
                'approve_workflow_test_app.view_approvetestmodel',
                'approve_workflow_test_app.add_relatedapprovetestmodel',
                'approve_workflow_test_app.approve_relatedapprovetestmodel',
                'approve_workflow_test_app.change_relatedapprovetestmodel',
                'approve_workflow_test_app.delete_relatedapprovetestmodel',
                'approve_workflow_test_app.view_relatedapprovetestmodel',
            )
        )
        self.client.force_login(user=minimal_test_user)

        # User sees the test app model:
        response = self.client.get(path='/admin/')
        self.assert_html_parts(response, parts=(
            '<title>Site administration | Django site admin</title>',
            '<a href="/admin/approve_workflow_test_app/approvetestmodel/">Approve test models</a>',
        ))

        url = (
            '/admin/approve_workflow_test_app/approvetestmodel'
            '/00000000-0000-0000-0000-000000000001/change/'
        )
        # User can change the entry and the approve button exists:
        response = self.client.get(url)
        assert response.status_code == 200
        self.assert_html_parts(response, parts=(
            '<input type="submit" value="Save and Approve" name="_approve" class="extra_submit">',
            'value="v1"',
            'value="relation one v1"',
        ))

        # We can save without the REQUIRED_FIELDS_PUBLIC fields:

        response = self.client.post(
            path=url,
            data={
                'name': 'v2',
                'title': '',
                'relations-TOTAL_FORMS': '1',
                'relations-INITIAL_FORMS': '1',
                'relations-MIN_NUM_FORMS': '0',
                'relations-MAX_NUM_FORMS': '1000',
                'relations-0-relation_name': 'relation one v2',
                'relations-0-relation_title': '',
                'relations-0-id': '00000000-0000-0000-0000-000000000002',
                'relations-0-main_entry': '00000000-0000-0000-0000-000000000001',
                'relations-__prefix__-relation_name': '',
                'relations-__prefix__-relation_title': '',
                'relations-__prefix__-id': '',
                'relations-__prefix__-main_entry': '00000000-0000-0000-0000-000000000001',
                '_save': 'Save',
            },
        )
        self.assertRedirects(
            response,
            expected_url='/admin/approve_workflow_test_app/approvetestmodel/',
            fetch_redirect_response=False
        )
        draft = ApproveTestModel.objects.get(pk=draft.pk)
        assert draft.name == 'v2'
        assert draft.approved is None
        relation_draft = RelatedApproveTestModel.objects.get(pk=relation_draft.pk)
        assert relation_draft.relation_name == 'relation one v2'
        assert relation_draft.approved is None

        # We can not approve without the REQUIRED_FIELDS_PUBLIC fields:

        response = self.client.post(
            path=url,
            data={
                'name': 'v3',
                'title': '',
                'relations-TOTAL_FORMS': '1',
                'relations-INITIAL_FORMS': '1',
                'relations-MIN_NUM_FORMS': '0',
                'relations-MAX_NUM_FORMS': '1000',
                'relations-0-relation_name': 'relation one v3',
                'relations-0-relation_title': '',
                'relations-0-id': '00000000-0000-0000-0000-000000000002',
                'relations-0-main_entry': '00000000-0000-0000-0000-000000000001',
                'relations-__prefix__-relation_name': '',
                'relations-__prefix__-relation_title': '',
                'relations-__prefix__-id': '',
                'relations-__prefix__-main_entry': '00000000-0000-0000-0000-000000000001',
                '_approve': 'Save+and+Approve',
            },
        )
        self.assert_html_parts(response, parts=(
            '<input type="submit" value="Save and Approve" name="_approve" class="extra_submit">',
            '<li>This field is required.</li>',
            'value="v3"',
            'value="relation one v3"',
        ))
        # Changes are not saved:
        draft = ApproveTestModel.objects.get(pk=draft.pk)
        assert draft.name == 'v2'
        assert draft.approved is None
        relation_draft = RelatedApproveTestModel.objects.get(pk=relation_draft.pk)
        assert relation_draft.relation_name == 'relation one v2'
        assert relation_draft.approved is None

        # We can approve if we add the REQUIRED_FIELDS_PUBLIC field values:

        with self.assertLogs('bx_django_utils', level=logging.INFO) as logs:
            response = self.client.post(
                path=url,
                data={
                    'name': 'v3',
                    'title': 'title now set',
                    'relations-TOTAL_FORMS': '1',
                    'relations-INITIAL_FORMS': '1',
                    'relations-MIN_NUM_FORMS': '0',
                    'relations-MAX_NUM_FORMS': '1000',
                    'relations-0-relation_name': 'relation one v3',
                    'relations-0-relation_title': 'relation title now set',
                    'relations-0-id': '00000000-0000-0000-0000-000000000002',
                    'relations-0-main_entry': '00000000-0000-0000-0000-000000000001',
                    'relations-__prefix__-relation_name': '',
                    'relations-__prefix__-relation_title': '',
                    'relations-__prefix__-id': '',
                    'relations-__prefix__-main_entry': '00000000-0000-0000-0000-000000000001',
                    '_approve': 'Save+and+Approve',
                },
            )
        logs_output = logs.output

        self.assertRedirects(
            response,
            expected_url='/admin/approve_workflow_test_app/approvetestmodel/',
            fetch_redirect_response=False
        )

        # Changes was saved and approve instances are made:

        draft = ApproveTestModel.objects.get(pk=draft.pk)
        approved = draft.approved
        assert tuple(
            ApproveTestModel.objects.values_list(
                'pk', 'name', 'title', 'is_draft'
            ).order_by('-is_draft')
        ) == (
            (draft.pk, 'v3', 'title now set', True),
            (approved.pk, 'v3', 'title now set', False)
        )
        assert tuple(
            RelatedApproveTestModel.objects.values_list(
                'main_entry_id', 'relation_name', 'relation_title', 'is_draft'
            ).order_by('-is_draft', 'relation_name')
        ) == (
            (draft.pk, 'relation one v3', 'relation title now set', True),
            (approved.pk, 'relation one v3', 'relation title now set', False),
        )
        assert approved is not None
        assert draft.name == approved.name == 'v3'
        assert draft.title == approved.title == 'title now set'

        rel_draft = RelatedApproveTestModel.objects.get(pk=relation_draft.pk)
        rel_approved = rel_draft.approved

        # assert rel_approved is not None
        assert rel_draft.relation_name == rel_approved.relation_name == 'relation one v3'
        assert rel_draft.relation_title == rel_approved.relation_title == 'relation title now set'

        assert logs_output == [
            "WARNING:bx_django_utils.approve_workflow.models:Not ready to approve:"
            " pk:UUID('00000000-0000-0000-0000-000000000001') model:ApproveTestModel",

            'INFO:bx_django_utils.approve_workflow.models:<RelatedApproveTestModel '
            'pk:00000000-0000-0000-0000-000000000002 (draft)> was approved to '
            f'<RelatedApproveTestModel pk:{rel_approved.pk} (approved)>',

            'INFO:bx_django_utils.approve_workflow.models:<ApproveTestModel '
            'pk:00000000-0000-0000-0000-000000000001 (draft, not ready)> was approved to '
            f'<ApproveTestModel pk:{approved.pk} (approved)>',
        ]

        messages = [m.message for m in get_messages(response.wsgi_request)]

        # "normalize" different Django versions:
        messages = [m.replace('“', '"').replace('”', '"') for m in messages]

        assert messages == [
            'The approve test model "<a href="/admin/approve_workflow_test_app/approvetestmodel'
            '/00000000-0000-0000-0000-000000000001/change/">&lt;ApproveTestModel '
            'pk:00000000-0000-0000-0000-000000000001 (draft, not ready)&gt;</a>" was '
            'changed successfully.',

            'Warning: "Ready to Approve" flag was not set!',

            f'<ApproveTestModel pk:{approved.pk} (approved)> was approved'
        ]



from django.conf import settings
from django.core.checks import Error, Warning, register
from django.utils.timezone import zoneinfo


MIDDLEWARE_STR = 'bx_django_utils.user_timezone.middleware.UserTimezoneMiddleware'


@register()
def user_timezone_check(app_configs, **kwargs):
    errors = []

    if MIDDLEWARE_STR not in settings.MIDDLEWARE:
        errors.append(
            Warning(
                msg='Missing UserTimezoneMiddleware',
                hint=f'Add "{MIDDLEWARE_STR}" to your settings.',
                id='user_timezone.W001',
            )
        )

    if not hasattr(settings, 'VISIBLE_TIMEZONES'):
        errors.append(
            Error(
                msg='Your settings has no "VISIBLE_TIMEZONES" defined!',
                hint="e.g.: VISIBLE_TIMEZONES = ['Europe/Berlin', 'America/Los_Angeles']",
                id='user_timezone.E001',
            )
        )
    else:
        for tz_name in settings.VISIBLE_TIMEZONES:
            try:
                zoneinfo.ZoneInfo(tz_name)
            except zoneinfo.ZoneInfoNotFoundError:
                errors.append(
                    Error(
                        msg=f'settings.VISIBLE_TIMEZONES entry: {tz_name!r} is invalid!',
                        id='user_timezone.E002',
                    )
                )

    return errors



from bx_py_utils.test_utils.datetime import parse_dt
from django.template import Context, Template
from django.test import SimpleTestCase
from django.utils import timezone, translation
from django.utils.timezone import zoneinfo

from bx_django_utils.test_utils.html_assertion import assert_html_snapshot


TEST_TEMPLATE = """\
{% load user_timezone %}
Datetime: {{ dt|humane_timezone_dt }}.
"""


class HumanizeTestCase(SimpleTestCase):
    def test_humane_timezone_dt(self):
        template = Template(TEST_TEMPLATE)

        with timezone.override(zoneinfo.ZoneInfo('UTC')), translation.override('en'):
            context = Context({'dt': parse_dt('2000-01-01T00:00:00+0000')})
            html = template.render(context)
        self.assertIn('Europe/Berlin: Jan. 1, 2000, 1 a.m.', html)
        self.assertIn('America/Los_Angeles: Dec. 31, 1999, 4 p.m.', html)
        self.assertIn('01/01/2000 midnight', html)
        self.assertInHTML('<small>(UTC)</small>', html)
        assert_html_snapshot(got=html, validate=False)

        with timezone.override(zoneinfo.ZoneInfo('Europe/Berlin')), translation.override('de-de'):
            context = Context({'dt': parse_dt('2000-01-01T00:00:00+0000')})
            html = template.render(context)
        self.assertIn('Europe/Berlin: 1. Januar 2000 01:00', html)
        self.assertIn('America/Los_Angeles: 31. Dezember 1999 16:00', html)
        self.assertIn('01.01.2000 01:00', html)
        self.assertInHTML('<small>(Europe/Berlin)</small>', html)
        assert_html_snapshot(got=html, validate=False)



from django.core.exceptions import ValidationError
from django.test import TestCase
from stdnum.exceptions import InvalidChecksum, InvalidFormat, InvalidLength

from bx_django_utils.data_types.gtin.constants import DEFAULT_ACCEPT_LENGTH
from bx_django_utils.data_types.gtin.form_fields import GtinFormField
from bx_django_utils.data_types.gtin.validators import GtinValidator, validate_gtin
from bx_django_utils_tests.test_app.models import GtinFieldTestModel


TEST_DATA = (
    '98412345678908',  # EAN-14
    '9783453059078',  # EAN-13 / ISBN-13
    '692771981161',  # UPC (12-digit)
    '3453059077',  # ISBN-10
    '73513537',  # EAN-8
)
NOT_VALID_TEST_DATA = (
    '10000000000000',  # EAN-14
    '1000000000000',  # EAN-13 / ISBN-13
    '100000000000',  # UPC (12-digit)
    '1000000000',  # ISBN-10
    '10000000',  # EAN-8
)


def iter_test_data():
    for number in TEST_DATA:
        length = len(number)
        yield (number, length)


def iter_not_valid_test_data():
    for number in NOT_VALID_TEST_DATA:
        length = len(number)
        yield (number, length)


class ValidatorTestCase(TestCase):
    def test_validate_gtin(self):
        for number, length in iter_test_data():
            validated_number = validate_gtin(number, accepted_length=(length,))
            assert validated_number == number

            with self.assertRaisesMessage(InvalidFormat, 'The number has an invalid format.'):
                validate_gtin('X' * length, accepted_length=(length,))

        err_msg = "The number's checksum or check digit is invalid."
        for number, length in iter_not_valid_test_data():
            with self.assertRaisesMessage(InvalidChecksum, err_msg):
                validate_gtin(number, accepted_length=(length,))

        for number, length in iter_test_data():
            if length in DEFAULT_ACCEPT_LENGTH:
                validated_number = validate_gtin(number)
                assert validated_number == number

            with self.assertRaisesMessage(InvalidLength, 'The number has an invalid length.'):
                validate_gtin(number='123')

    def test_GtinValidator(self):
        for number, length in iter_test_data():
            validator = GtinValidator(accepted_length=(length,))
            validator(number)

        default_validator = GtinValidator()
        for number, length in iter_test_data():
            if length in DEFAULT_ACCEPT_LENGTH:
                default_validator(number)

    def test_validator_eq(self):
        validator1 = GtinValidator()
        validator2 = GtinValidator(accepted_length=(8,))
        validator3 = GtinValidator(accepted_length=DEFAULT_ACCEPT_LENGTH)
        assert validator1 != validator2
        assert validator1 == validator3


class GtinFormFieldTestCase(TestCase):
    def test_basic(self):
        for number, length in iter_test_data():
            field = GtinFormField(accepted_length=(length,))
            field.clean(number)
            with self.assertRaisesMessage(ValidationError, 'The number has an invalid format.'):
                field.clean('X' * length)

        err_msg = 'The number\'s checksum or check digit is invalid.'
        field = GtinFormField(accepted_length=(8, 10, 12, 13, 14))
        for number in NOT_VALID_TEST_DATA:
            with self.assertRaisesMessage(ValidationError, err_msg):
                field.clean(number)

        assert field.prepare_value('978-3-453-05907-9') == '9783453059079'


class GtinFieldTestModelTestCase(TestCase):
    def test_basic(self):
        instance = GtinFieldTestModel.objects.create(
            default_gtin='692771981161',  # UPC (12-digit)
            all_gtin='98412345678908',  # EAN-14
            ean13='9783453059078',  # EAN-13 / ISBN-13
        )
        instance.full_clean()

        # Our model field used our form field?
        model_field = instance._meta.get_field('default_gtin')
        form_field = model_field.formfield()
        assert isinstance(form_field, GtinFormField)

        err_msg = 'The number has an invalid format.'
        for number, length in iter_test_data():
            instance.all_gtin = number
            if length in DEFAULT_ACCEPT_LENGTH:
                instance.default_gtin = number
            if length == 13:
                instance.ean13 = number
            instance.full_clean()

            instance.all_gtin = 'X' * length
            with self.assertRaisesMessage(ValidationError, err_msg):
                instance.full_clean()

        err_msg = '{\'all_gtin\': ["The number\'s checksum or check digit is invalid."]}'
        for number in NOT_VALID_TEST_DATA:
            instance.all_gtin = number
            with self.assertRaisesMessage(ValidationError, err_msg):
                instance.full_clean()

    def test_none_value(self):
        instance = GtinFieldTestModel.objects.create(
            default_gtin='692771981161',  # required
            all_gtin='',  # blank=True, null=False
            ean13=None,  # null=True, blank=True,
        )
        instance.full_clean()

        all_gtin_form_field = instance._meta.get_field('all_gtin').formfield()
        assert all_gtin_form_field.clean(None) == ''
        assert all_gtin_form_field.clean('') == ''
        assert all_gtin_form_field.prepare_value(None) is None
        assert all_gtin_form_field.prepare_value('') == ''

        ean13_form_field = instance._meta.get_field('ean13').formfield()
        assert ean13_form_field.clean(None) is None
        assert ean13_form_field.clean('') is None
        assert ean13_form_field.prepare_value(None) is None
        assert ean13_form_field.prepare_value('') == ''



from django import template

from bx_django_utils.user_timezone.humanize import human_timezone_datetime


register = template.Library()


@register.filter
def humane_timezone_dt(value):
    """
    Template filter to render a datetime with timezone information.
    """
    return human_timezone_datetime(dt=value)



import io
import tempfile
from pathlib import Path

from bx_py_utils.test_utils.snapshot import assert_text_snapshot
from django.core.management import call_command
from django.test import SimpleTestCase, TestCase

from bx_django_utils.test_utils.fixtures import (
    FIXTURES_FILE_PATHS,
    BaseFixtures,
    FixturesRegistry,
    SerializerFixtures,
    fixtures_registry,
)
from bx_django_utils_tests.test_app.management.commands import renew_fixtures
from bx_django_utils_tests.test_app.models import ColorFieldTestModel
from bx_django_utils_tests.test_app.tests.fixtures.fixtures1 import (
    EXAMPLE_FIXTURES_DATA1,
    EXAMPLE_FIXTURES_DATA2,
    ExampleFixtures1,
    ExampleFixtures2,
)


class FixturesHelperTestCase(SimpleTestCase):
    def reset(self):
        FIXTURES_FILE_PATHS.clear()

    def setUp(self):
        super().setUp()
        self.reset()

    def tearDown(self):
        super().tearDown()
        self.reset()

    def test_example_fixtures(self):
        assert ExampleFixtures1().get_fixture_data() == EXAMPLE_FIXTURES_DATA1
        assert ExampleFixtures2().get_fixture_data() == EXAMPLE_FIXTURES_DATA2

    def test_registry(self):
        assert sorted(fixtures_registry.fixtures.keys()) == ['ExampleFixtures1', 'ExampleFixtures2']

    def test_renew(self):
        fixture = ExampleFixtures1()

        # Change fixtures file content:
        fixture.store_fixture_data(data=['something else!'])
        assert fixture.get_fixture_data() == ['something else!']

        # "renew" -> recreate json dump back:
        fixture.renew()

        # The "old" data?
        assert fixture.get_fixture_data() == EXAMPLE_FIXTURES_DATA1

    def test_renew_command(self):
        # Change fixtures file content:
        ExampleFixtures1().store_fixture_data(data=['something else!'])
        assert ExampleFixtures1().get_fixture_data() == ['something else!']

        # Call "renew" command:
        capture_stdout = io.StringIO()
        capture_stderr = io.StringIO()
        call_command(renew_fixtures.Command(), stdout=capture_stdout, stderr=capture_stderr)
        stdout_output = capture_stdout.getvalue()
        stderr_output = capture_stderr.getvalue()
        assert '2 Fixtures updated, ok.' in stdout_output
        assert stderr_output == ''
        assert_text_snapshot(got=stdout_output)

        assert ExampleFixtures1().get_fixture_data() == EXAMPLE_FIXTURES_DATA1

    def test_overwriting(self):
        class Foo(BaseFixtures):
            base_path = Path(__file__).parent
            file_name = 'bar.json'

        class Bar(Foo):
            pass

        registry = FixturesRegistry()
        registry.register()(Foo)

        # It's not possible to register more than two times:
        with self.assertRaisesMessage(RuntimeError, 'Fixture class "Foo" already registered!'):
            registry.register()(Foo)

        foo = registry.fixtures['Foo']
        file_path = foo.file_path

        # It's not possible to overwrite the same JSON file:
        with self.assertRaisesMessage(
            RuntimeError, f'File path "{file_path}" from "Bar" already used!'
        ):
            registry.register()(Bar)


class SerializerFixturesTestCase(TestCase):
    def test_basic(self):
        with tempfile.NamedTemporaryFile(prefix='serializer_fixtures_', suffix='.json') as tmp_file:
            tmp_file_path = Path(tmp_file.name)

            class Example(SerializerFixtures):
                """Just a test example implementation"""

                base_path = tmp_file_path.parent
                file_name = tmp_file_path.name

                def renew(self):
                    ColorFieldTestModel.objects.create(pk=1, required_color='#00AAff')
                    qs = ColorFieldTestModel.objects.all()
                    self.store_fixture_data(queryset=qs)

            example = Example()

            # Store the model instance into the JSON file:
            example.renew()

            json_str = tmp_file_path.read_text()
            self.assertIn('#00AAff', json_str)

            ColorFieldTestModel.objects.all().delete()
            self.assertEqual(ColorFieldTestModel.objects.count(), 0)

            # Maybe useful: It's possible to get the serializer data:
            data = example.get_fixture_data()
            self.assertEqual(
                data,
                [
                    {
                        'fields': {'optional_color': None, 'required_color': '#00AAff'},
                        'model': 'test_app.colorfieldtestmodel',
                        'pk': 1,
                    }
                ],
            )

            # Recreate the objects:
            instances = example.create_objects()
            self.assertEqual(ColorFieldTestModel.objects.count(), 1)
            created_instance = ColorFieldTestModel.objects.first()
            self.assertEqual(instances, [created_instance])



from django.http import HttpResponse
from lxml import etree


def _sorted_join_repr(iterable):  # TODO: Move to bx_py_utils?
    """
    >>> _sorted_join_repr(['baz', 'foo', 'bar'])
    "'bar', 'baz', 'foo'"
    """
    return ', '.join(repr(item) for item in sorted(iterable))


class AssertFormFields:
    """
    Helper to check the existing form fields.
    e.g.:
        response = self.client.get('/foo/bar/form.html')
        assert_form_fields = AssertFormFields(response)
        self.assertGreaterEqual(len(assert_form_fields), 30) # Min. 30 fields?
        assert_form_fields.assert_field_names_not_exists(
            field_names={'internal', 'foo', 'bar'}
        )
        assert_form_fields.assert_field_names_exists(
            field_names={'normal_field1', 'normal_field2'}
        )
        assert_snapshot(got=assert_form_fields.data)
    """

    def __init__(
        self,
        response: HttpResponse,
        xpath_prefix='//div[@id="content"]',  # Collect fields only in this page part
        xpath_filter='[not(@type="hidden")]',  # Ignore hidden fields as default
    ):

        parser = etree.HTMLParser()
        tree = etree.fromstring(response.content, parser=parser)
        form_fields = tree.xpath(
            f'{xpath_prefix}//form//*[self::input or self::textarea or self::select]{xpath_filter}'
        )

        # Collect only the attributes of the <input> field:
        self.data = []
        for element in form_fields:
            attrib = dict(element.attrib)  # Make if JSON serializable for snapshot test
            self.data.append(attrib)

        # The main purpose here is to test which fields are available,
        # not how they are sorted on the website.
        #
        # Sort them for snapshot test:
        self.data.sort(key=lambda x: (x.get('name', ''), x.get('id', '')))

    def get_all_field_names(self) -> set:
        """
        :return: All existing field names (NOTE: Ignores fields without name!)
        """
        all_field_names = {attrib['name'] for attrib in self.data if attrib.get('name')}
        return all_field_names

    def assert_field_names_exists(self, field_names) -> None:
        """
        Check that form field exists by name
        """
        expected_field_names = set(field_names)
        existing_field_names = self.get_all_field_names()
        missing_field_names = expected_field_names - existing_field_names
        assert not missing_field_names, (
            f'Fields: {_sorted_join_repr(missing_field_names)}'
            f' are not present in: {_sorted_join_repr(existing_field_names)}'
        )

    def assert_field_names_not_exists(self, field_names) -> None:
        """
        Check that form fields doesn't exist by name
        """
        should_not_present = set(field_names)
        existing_field_names = self.get_all_field_names()
        occurring_names = should_not_present & existing_field_names
        assert not occurring_names, (
            f'These fields should not occur,'
            f' but are present: {_sorted_join_repr(occurring_names)}'
        )

    def __len__(self):
        return len(self.data)



from django.contrib import admin

from bx_django_utils.admin_utils.filters import ExistingCountedListFilter
from bx_django_utils_tests.test_app.admin_filters_demo import NotAllSimpleListFilterDemo
from bx_django_utils_tests.test_app.models import ColorFieldTestModel, CreateOrUpdateTestModel


class NameListFilter(ExistingCountedListFilter):
    title = 'name'
    parameter_name = 'name'
    model_field_name = 'name'


class SlugListFilter(ExistingCountedListFilter):
    title = 'slug'
    parameter_name = 'slug'
    model_field_name = 'slug'


@admin.register(CreateOrUpdateTestModel)
class CreateOrUpdateTestModelAdmin(admin.ModelAdmin):
    list_display = ('human_create_dt', 'human_update_dt', 'name', 'slug')
    list_display_links = ('name', 'slug')
    readonly_fields = ('create_dt', 'update_dt')
    prepopulated_fields = {'slug': ('name',)}
    list_filter = (NotAllSimpleListFilterDemo, NameListFilter, SlugListFilter)


@admin.register(ColorFieldTestModel)
class ColorFieldTestModelAdmin(admin.ModelAdmin):
    list_display = ('required_color', 'optional_color')



"""
    Helper to publish this Project to PyPi
"""

from pathlib import Path

from poetry_publish.publish import poetry_publish
from poetry_publish.utils.subprocess_utils import verbose_check_call

import bx_django_utils


PACKAGE_ROOT = Path(bx_django_utils.__file__).parent.parent


def publish():
    """
        Publish to PyPi
        Call this via:
            $ poetry run publish
    """
    verbose_check_call('make', 'test')  # don't publish if tests fail
    verbose_check_call('make', 'fix-code-style')  # don't publish if code style wrong

    poetry_publish(package_root=PACKAGE_ROOT, version=bx_django_utils.__version__)



from django.views.generic import RedirectView

from bx_django_utils.admin_extra_views.base_view import AdminExtraViewMixin
from bx_django_utils.admin_extra_views.utils import reverse_admin_extra_view


class Redirect2AdminExtraView(RedirectView):
    """
    Redirect to a Admin Extra Views.
    """

    admin_view = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        assert self.admin_view is not None
        if not issubclass(self.admin_view, AdminExtraViewMixin):
            raise ValueError(f'Admin view {self.admin_view} must subclass AdminExtraViewMixin.')

    def get_redirect_url(self, *args, **kwargs):
        url = reverse_admin_extra_view(view_class=self.admin_view)
        return url



from django.contrib import admin

from bx_django_utils.approve_workflow.admin import BaseApproveModelAdmin
from bx_django_utils.approve_workflow.forms import PublishAdminForm
from bx_django_utils_tests.approve_workflow_test_app.models import ApproveTestModel, RelatedApproveTestModel


class RelatedApproveTestModelInline(admin.StackedInline):
    form = PublishAdminForm  # Activate models REQUIRED_FIELDS_PUBLIC on approve
    model = RelatedApproveTestModel
    exclude = ('blocked',)
    extra = 0


@admin.register(ApproveTestModel)
class ApproveTestModelAdmin(BaseApproveModelAdmin):
    inlines = (RelatedApproveTestModelInline,)



from django.contrib.auth.models import User
from playwright.sync_api import BrowserContext, expect

from bx_django_utils.test_utils.playwright import PlaywrightTestCase


class TestPlaywrightTestCase(PlaywrightTestCase):

    def setUp(self):
        super().setUp()
        self.superuser = User.objects.create_superuser(
            username='a-superuser',
            password='ThisIsNotAPassword!',
        )

    def test_login(self):
        context: BrowserContext = self.browser.new_context(
            ignore_https_errors=True,  # StaticLiveServerTestCase has no valid certificate ;)
            locale='en-US',
        )
        with context.new_page() as page:
            # Redirect to login?
            page.goto(f'{self.live_server_url}/admin/')
            expect(page).to_have_url(f'{self.live_server_url}/admin/login/?next=/admin/')
            expect(page).to_have_title('Log in | Django site admin')

            # Login:
            page.fill('#id_username', 'a-superuser')
            page.fill('#id_password', 'ThisIsNotAPassword!')
            page.keyboard.press('Enter')

            # Are we logged in?
            expect(page).to_have_url(f'{self.live_server_url}/admin/')
            expect(page).to_have_title('Site administration | Django site admin')


class TestWithFirefoxTestCase(PlaywrightTestCase):
    BROWSER_NAME = 'firefox'

    def test_browser(self):
        self.assertEqual(self.browser_name, 'firefox')
        self.assertEqual(self.browser_type.name, 'firefox')


class TestWithChromiumTestCase(PlaywrightTestCase):
    BROWSER_NAME = 'chromium'

    def test_browser(self):
        self.assertEqual(self.browser_name, 'chromium')
        self.assertEqual(self.browser_type.name, 'chromium')



import datetime

from django.template import Library
from django.utils import dateformat
from django.utils.formats import get_format
from django.utils.html import format_html
from django.utils.timezone import is_aware

from bx_django_utils.humanize.time import human_timedelta


register = Library()


@register.filter
def human_duration(value, arg=None):
    """
    Verbose time since template tag, e.g.: `<span title="Jan. 1, 2000, noon">2.0 seconds</span>`

    e.g.:
        <span title="Jan. 1, 2000, noon">2.0 seconds</span>
    """
    if not value:
        return ''

    try:
        # Convert datetime.date to datetime.datetime for comparison.
        if not isinstance(value, datetime.datetime):
            value = datetime.datetime(value.year, value.month, value.day)

        if arg and not isinstance(arg, datetime.datetime):
            arg = datetime.datetime(arg.year, arg.month, arg.day)

        arg = arg or datetime.datetime.now(datetime.timezone.utc if is_aware(value) else None)

        delta_str = human_timedelta(t=arg - value)
        time_str = dateformat.format(value, get_format('DATETIME_FORMAT'))
    except (AttributeError, ValueError, TypeError):
        return ''

    return format_html(
        '<span title="{time_str}">{delta_str}</span>',
        time_str=time_str,
        delta_str=delta_str,
    )



import logging

from django.contrib import admin, messages
from django.contrib.auth import get_permission_codename
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from bx_django_utils.approve_workflow.constants import POST_KEY_APPROVE
from bx_django_utils.approve_workflow.forms import PublishAdminForm


logger = logging.getLogger(__name__)


class BaseApproveModelAdmin(admin.ModelAdmin):
    """
    Base admin class for a draft/approve Model
    """
    form = PublishAdminForm  # Activate models REQUIRED_FIELDS_PUBLIC on approve
    change_form_template = 'approve_workflow/change_form.html'
    readonly_fields = (
        'approved',
        'is_draft', 'ready_to_approve',
        'create_dt', 'update_dt'
    )
    list_display = (
        '__str__',
        'approved',
        'is_draft', 'ready_to_approve',
        'create_dt', 'update_dt'
    )

    def has_approve_permissions(self, user):
        opts = self.opts
        codename = get_permission_codename('approve', opts)
        perm = f'{opts.app_label}.{codename}'
        has_permission = user.has_perm(perm)
        if not has_permission:
            logger.debug('User %r has not %r permissions', user.pk, perm)
        return has_permission

    def render_change_form(self, request, context, obj=None, **kwargs):
        if obj is None or not obj.is_draft:
            show_approve = False
        else:
            # Display "Approve" button in submit line if user has approve permissions
            show_approve = self.has_approve_permissions(user=request.user)

        if obj is None or obj.is_draft:
            show_approve_warning = False
        else:
            # Use changed the approved version -> Display warning about this
            show_approve_warning = True

        context.update({
            'show_approve_warning': show_approve_warning,
            'show_approve': show_approve,
        })
        return super().render_change_form(request, context, **kwargs)

    def get_form(self, request, obj=None, **kwargs):
        Form = super().get_form(request, obj, **kwargs)

        # Override form when pressing approve button
        if POST_KEY_APPROVE in request.POST and obj:
            class ApproveCheckForm(Form):
                def clean(self):
                    if self.cleaned_data['blocked']:
                        raise ValidationError({'blocked': 'Blocked item cannot be approved!'})
                    return super().clean()

            return ApproveCheckForm

        return Form

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)

        if POST_KEY_APPROVE in request.POST:
            # save instance, because we can only approve after relations are saved, too.
            logger.debug('to_approve=%r', obj)
            self.to_approve = obj

    def _approve(self, request, obj):
        can_approve = self.has_approve_permissions(user=request.user)
        if not can_approve:
            raise PermissionError('User can not approve')

        ready_to_approve = obj.ready_to_approve
        approved = obj.approve()
        logger.debug('%r approved to %r', obj, approved)
        if not ready_to_approve:
            messages.warning(request, _('Warning: "Ready to Approve" flag was not set!'))
        self.message_user(request, _('%s was approved') % approved, messages.SUCCESS)

    def _changeform_view(self, request, object_id, form_url, extra_context):
        """
        It's a little bit tricky to approve a instance, because we can only
        approve a instance after it was saved and after all relations was saved!
        """
        self.to_approve = None
        response = super()._changeform_view(request, object_id, form_url, extra_context)
        if self.to_approve is not None:
            # model and all relations was saved and self.save_model() has set self.to_approve
            self._approve(request, obj=self.to_approve)

        return response



import json
import pathlib
import tempfile
from unittest import mock

import django
from bx_py_utils.compat import removesuffix
from bx_py_utils.environ import OverrideEnviron
from bx_py_utils.test_utils.snapshot import _get_caller_names
from django.contrib import messages
from django.contrib.messages.storage.session import SessionStorage
from django.template.defaulttags import CsrfTokenNode
from django.test import RequestFactory, SimpleTestCase

from bx_django_utils.test_utils.html_assertion import (
    HtmlAssertionMixin,
    assert_html_response_snapshot,
)


class FakeResponse:
    def __init__(self, request):
        self.wsgi_request = request


def get_request():
    request = RequestFactory().request()
    request.session = {}
    request._messages = SessionStorage(request)
    return request


class HtmlAssertionTestCase(HtmlAssertionMixin, SimpleTestCase):
    def test_assert_messages(self):
        request = get_request()
        messages.info(request, 'First message.')
        messages.warning(request, 'Second message.')
        messages.error(request, 'The last message.')
        response = FakeResponse(request)
        self.assert_messages(response, expected_messages=[
            'First message.',
            'Second message.',
            'The last message.'
        ])

        msg = (
            'Messages are not equal:\n'
            '--- got\n'
            '\n'
            '+++ expected\n'
            '\n'
            '@@ -1,5 +1,5 @@\n'
            '\n'
            ' [\n'
            '     "First message.",\n'
            '-    "Second message.",\n'
            '+    "Second X message.",\n'
            '     "The last message."\n'
            ' ]'
        )
        with self.assertRaisesMessage(AssertionError, msg):
            self.assert_messages(response, expected_messages=[
                'First message.',
                'Second X message.',
                'The last message.'
            ])

    def test_snapshot_messages(self):
        request = get_request()
        messages.info(request, 'First message.')
        messages.warning(request, 'Second message.')
        messages.error(request, 'The last message.')
        response = FakeResponse(request)

        with tempfile.TemporaryDirectory() as tmp_dir, OverrideEnviron(
            RAISE_SNAPSHOT_ERRORS='1'  # Maybe it's disabled in this test run!
        ), self.assertRaises(FileNotFoundError):
            self.snapshot_messages(response, root_dir=tmp_dir)

            # Test that snapshot got written
            snapshot_name = 'test_html_assertion_snapshot_messages_1'
            snapshot_file = (pathlib.Path(tmp_dir) / f'{snapshot_name}.snapshot.json')
            snapshot_content = json.loads(snapshot_file.read_text())
            self.assertEqual(
                snapshot_content, ['First message.', 'Second message.', 'The last message.'])

            # Should work now
            self.snapshot_messages(response, root_dir=tmp_dir, snapshot_name=snapshot_name)

    def test_assert_html_response_snapshot(self):
        # Add the Django version to snapshot name, because they have different outputs ;)
        root_dir, snapshot_name = _get_caller_names()
        snapshot_name = removesuffix(snapshot_name, '.snapshot')
        version = django.__version__[:3]
        snapshot_name += f'-django{version}'

        with mock.patch.object(CsrfTokenNode, 'render', return_value='MockedCsrfTokenNode'):
            assert_html_response_snapshot(
                response=self.client.get(path='/admin/login/'),
                status_code=200,
                root_dir=root_dir,
                snapshot_name=snapshot_name,
                validate=False,
                query_selector=None
            )

        response = self.client.get(path='/admin/')
        self.assertRedirects(
            response, expected_url='/admin/login/?next=%2Fadmin%2F', fetch_redirect_response=False
        )

        # A redirect has no content, so no snapshot file will be created,
        # but the status code is checked, too:
        assert_html_response_snapshot(
            response, status_code=302, validate=False, query_selector=None)

        msg = 'Status code is 302 but excepted 200'
        with self.assertRaisesMessage(AssertionError, msg):
            assert_html_response_snapshot(response, validate=False, query_selector=None)



"""
    Activate "ExtraViewAdminSite" by set this as default admin site
"""
from django.contrib.admin.apps import AdminConfig


class CustomAdminConfig(AdminConfig):
    """
    Change Django Admin Site to ExtraViewAdminSite for the extra views.
    """

    default_site = 'bx_django_utils.admin_extra_views.site.ExtraViewAdminSite'



"""
    Utilities to manage text fixtures in JSON files.
    A registry to collect all fixtures and a manage command to "renew" all fixtures.

    Usage, see:
     * bx_django_utils_tests/test_app/tests/fixtures/fixtures1.py
     * bx_django_utils_tests/test_app/management/commands/renew_fixtures.py
"""
import json
import os
import re
import sys
from importlib import import_module
from pathlib import Path

from bx_py_utils.path import assert_is_dir, assert_is_file
from django.apps import apps
from django.core import serializers
from django.core.management import BaseCommand
from django.core.serializers.base import DeserializedObject

from bx_django_utils.json_utils import to_json


FIXTURES_FILE_PATHS = set()


class BaseFixtures:
    """
    Base class for JSON dump fixtures.
    """

    base_path = None
    file_name = None

    def __init__(self):
        assert self.base_path, 'No base_path defined!'
        assert self.file_name, 'No file_name defined!'
        assert self.file_name.endswith('.json'), f'File extension not ".json": {self.file_name!r}'
        assert_is_dir(self.base_path)
        self.file_path = self.base_path / self.file_name

    def get_fixture_data(self):
        try:
            assert_is_file(self.file_path)
        except OSError as err:
            raise OSError(f'{err} (Hint: renew test fixtures!)')
        with Path(self.file_path).open('r') as f:
            return json.load(f)

    def store_fixture_data(self, data):
        assert data, 'Given data is empty!'
        with Path(self.file_path).open('w') as f:
            f.write(to_json(data, indent=2))

    def renew(self):
        """
        Endpoint to update the existing fixture data.
        Should be use self.store_fixture_data()
        """
        raise NotImplementedError


class SerializerFixtures(BaseFixtures):
    """
    Helper to store/restore model instances serialized into a JSON file.
    """

    def store_fixture_data(self, queryset):
        """
        Save the given QuerySet data into JSON fixture file
        """
        assert queryset.count() >= 1
        with Path(self.file_path).open('w') as f:
            serializers.serialize(
                format='json',
                queryset=queryset,
                indent=2,
                sort_keys=True,
                stream=f,
            )

    def create_objects(self):
        """
        Create model entries from stored JSON fixtures.
        returns the created model instance.
        """
        instances = []
        with Path(self.file_path).open('r') as f:
            for unsaved_instance in serializers.deserialize('json', f):
                assert isinstance(unsaved_instance, DeserializedObject)
                unsaved_instance.object.full_clean()
                unsaved_instance.save()
                instances.append(unsaved_instance.object)
        return instances


class FixturesRegistry:
    """
    Registry to collect a list of all existing fixture classes.
    """

    def __init__(self):
        self.fixtures = {}

    def register(self):
        def _class_wrapper(FixtureClass):
            assert issubclass(FixtureClass, BaseFixtures)

            class_name = FixtureClass.__name__

            if class_name in self.fixtures:
                raise RuntimeError(f'Fixture class "{class_name}" already registered!')

            fixtures = FixtureClass()
            file_path = fixtures.file_path
            if file_path in FIXTURES_FILE_PATHS:
                raise RuntimeError(f'File path "{file_path}" from "{class_name}" already used!')
            FIXTURES_FILE_PATHS.add(file_path)

            self.fixtures[class_name] = fixtures
            return FixtureClass

        return _class_wrapper

    def __iter__(self):
        for class_name, fixtures in self.fixtures.items():
            yield fixtures


fixtures_registry = FixturesRegistry()


def autodiscover(verbosity) -> int:
    """
    Register all fixtures by import all **/fixtures/**/*.py files
    """

    def get_fixtures(path):
        for root, dirs, files in os.walk(path, followlinks=False):
            if 'fixtures' in dirs:
                yield from Path(root, 'fixtures').glob('**/*.py')

    count = 0
    for app_config in apps.get_app_configs():
        path = app_config.path

        for fixtures_path in get_fixtures(path):
            rel_path = fixtures_path.relative_to(path)
            parts = rel_path.with_suffix('').parts
            rel_pkg = '.'.join(parts)
            full_pkg = f'{app_config.name}.{rel_pkg}'
            if verbosity > 1:
                print(f'Fixtures found: {full_pkg!r}')
            import_module(full_pkg)
            count += 1
    return count


class RenewAllFixturesBaseCommand(BaseCommand):
    """
    A base Django manage command to renew all existing fixture JSON dump files
    """

    help = 'Renew all (example) test fixtures'

    def add_arguments(self, parser):
        parser.add_argument(
            '-f',
            '--filter',
            metavar='REGEX',
            help='Filter fixture class names or files by a regular expression',
        )

    def handle(self, *args, **options):
        verbosity = options['verbosity']

        count = autodiscover(verbosity)  # fill registry by import all fixture files
        if verbosity:
            self.stdout.write(f'Found {count} fixtures files.')

        if count < 1:
            self.stderr.write('No fixtures files found in all installed Django apps!')
            sys.exit(1)

        if not fixtures_registry.fixtures:
            self.stderr.write('No fixtures class registered!')
            sys.exit(1)

        all_fixtures = list(fixtures_registry)

        if filter_re := options['filter']:
            filter_rex = re.compile(filter_re)
            all_fixtures = [
                f
                for f in all_fixtures
                if filter_rex.search(f.__class__.__name__) or filter_rex.search(f.file_name)
            ]

        no = 0
        for no, fixture in enumerate(all_fixtures, 1):
            if verbosity:
                self.stdout.write('_' * 100)
                self.stdout.write(
                    f'{no}. renew "{fixture.__class__.__name__}" file "{fixture.file_name}...'
                )
            fixture.renew()

        if verbosity:
            self.stdout.write(f'\n{no} Fixtures updated, ok.\n')



import inspect
import pickle
from collections import OrderedDict
from threading import Lock
from unittest import mock

from django.core.cache import cache
from django.core.cache.backends.base import default_key_func
from django.core.cache.backends.locmem import LocMemCache


class FakeCache(LocMemCache):

    def __init__(self):
        # BaseCache attributes:
        self.default_timeout = None
        self._max_entries = 300
        self._cull_frequency = 3
        self.key_prefix = ''
        self.version = 1
        self.key_func = default_key_func

        # LocMemCache attributes:
        self._cache = OrderedDict()
        self._expire_info = {}
        self._lock = Lock()

        # Storage for 'clean' cache data, without version:
        self.data = {}

    def add(self, key, value, **kwargs):
        if key not in self.data:
            self.data[key] = value
        return super().add(key, value, **kwargs)

    def set(self, key, value, **kwargs):
        self.data[key] = value
        return super().set(key, value, **kwargs)

    def delete(self, key, **kwargs):
        if key in self.data:
            del self.data[key]
        return super().delete(key, **kwargs)

    def incr(self, key, *args, **kwargs):
        new_value = super().incr(key, *args, **kwargs)
        self.data[key] = new_value
        return new_value

    def clear(self):
        self.data.clear()
        super().clear()


class MockCache:
    """
    Mock Django cache backend, so it's easy to check/manipulate the cache content
    """

    def __init__(self):
        self.fake_cache = FakeCache()

        self.mocks = []
        methods = inspect.getmembers(self.fake_cache, inspect.ismethod)
        for name, func in methods:
            if name.startswith('_'):
                continue

            self.mocks.append(
                mock.patch(f'django.core.cache.cache.{name}', func)
            )

    @property
    def data(self):
        """
        :return: "cleaned" cache data, without version
        """
        return self.fake_cache.data

    @property
    def raw_data(self):
        """
        :return: cache data with all versions
        """
        return {
            key: pickle.loads(pickled)
            for key, pickled in self.fake_cache._cache.items()
        }

    def __enter__(self):
        for m in self.mocks:
            m.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        for m in self.mocks:
            m.__exit__(exc_type, exc_value, traceback)


class ClearCacheMixin:
    """
    TestCase mixin to clear the Django cache in setUp/tearDown
    """

    def setUp(self):
        super().setUp()
        cache.clear()

    def tearDown(self):
        super().tearDown()
        cache.clear()



from django.test import TestCase

from bx_django_utils.test_utils.model_clean_assert import AssertModelCleanCalled
from bx_django_utils_tests.test_app.models import CreateOrUpdateTestModel


class ModelCleanAssertTestCase(TestCase):
    def test_full_clean_called(self):
        with AssertModelCleanCalled() as cm:
            instance = CreateOrUpdateTestModel(name='foo', slug='bar')
            instance.full_clean()
            instance.save()

        cm.assert_no_missing_cleans()
        assert cm.missing_cleans == []
        assert len(cm.called_cleans) == 1
        clean_mock = cm.called_cleans[0]
        assert repr(clean_mock) == '<CleanMock test_app.CreateOrUpdateTestModel>'
        assert clean_mock.clean_called is True

    def test_full_clean_not_called(self):
        with AssertModelCleanCalled() as cm:
            instance = CreateOrUpdateTestModel(name='foo', slug='bar')
            instance.save()

        assert cm.called_cleans == []
        assert len(cm.missing_cleans) == 1
        clean_mock = cm.missing_cleans[0]
        assert repr(clean_mock) == '<CleanMock test_app.CreateOrUpdateTestModel>'
        assert clean_mock.clean_called is False

        with self.assertRaises(AssertionError) as err:
            cm.assert_no_missing_cleans()

        error_msg = str(err.exception)
        assert 'There are 1 missing clean calls:' in error_msg
        assert 'test_app.CreateOrUpdateTestModel:' in error_msg

        # The stack:
        assert __file__ in error_msg
        assert 'in test_full_clean_not_called' in error_msg



from typing import Iterator

from django.urls import reverse

from bx_django_utils.admin_extra_views.base_view import AdminExtraViewMixin
from bx_django_utils.admin_extra_views.datatypes import AdminExtraMeta
from bx_django_utils.admin_extra_views.registry import extra_view_registry


def reverse_admin_extra_view(view_class) -> str:
    """
    Get the URL of a Admin Extra View, e.g.: url=reverse_admin_extra_view(YouAdminExtraView)
    """
    assert issubclass(view_class, AdminExtraViewMixin)
    assert isinstance(view_class.meta, AdminExtraMeta)
    url = reverse(view_class.meta.url_name)
    return url


def iter_admin_extra_views_urls() -> Iterator[str]:
    """
    Iterate over all registered admin extra view urls.
    """
    for view_class in extra_view_registry:
        url = reverse_admin_extra_view(view_class)
        yield url



import datetime


class MockDatetimeGenerator:
    """
    Mock django `timezone.now()` with generic time stamps in tests.

    Note: Set a lower offset if you use the Django test client!
          By default every timezone.now() will increase by 1 Year,
          so a session is practically always expired ;)

    e.g.:

    class FooBar(TestCase):
        @mock.patch.object(timezone, 'now', MockDatetimeGenerator())
        def test_something_without_test_client(self):
            ...

    or:

    def test_foobar():
        offset = datetime.timedelta(minutes=1)
        with mock.patch.object(timezone, 'now', MockDatetimeGenerator(offset=offset)):
            ...
    """

    def __init__(self, offset=None):
        if offset is not None:
            assert isinstance(offset, datetime.timedelta)
            self.now = datetime.datetime(
                2000, 1, 1, 0, 0, 0,
                tzinfo=datetime.timezone.utc
            )
        else:
            self.now = None

        self.offset = offset
        self.num = 0

    def __call__(self):
        if self.offset is None:
            self.num += 1
            dt = datetime.datetime(
                2000 + self.num, 1, 1, 0, 0, 0,
                tzinfo=datetime.timezone.utc
            )
            return dt
        else:
            self.now += self.offset
            return self.now



from django.contrib.auth.models import Group, User
from django.core.exceptions import FieldError
from django.db.models import Q
from django.test import TestCase

from bx_django_utils.models.queryset_utils import remove_filter, remove_model_filter


class QuerySetUtilsTestCase(TestCase):
    def test_remove_filter_basic(self):
        User.objects.create(username='bar1')
        User.objects.create(username='bar2')
        User.objects.create(username='foo1')
        User.objects.create(username='foo2')

        queryset = User.objects.order_by('username').all()
        assert list(queryset.values_list('username', flat=True)) == [
            'bar1', 'bar2', 'foo1', 'foo2'
        ]

        queryset = queryset.filter(username='foo1')
        assert list(queryset.values_list('username', flat=True)) == ['foo1']

        old_id = id(queryset)
        queryset = remove_filter(queryset, lookup='username')
        assert list(queryset.values_list('username', flat=True)) == [
            'bar1', 'bar2', 'foo1', 'foo2'
        ]
        assert id(queryset) != old_id

        queryset = queryset.filter(username__startswith='bar')
        assert list(queryset.values_list('username', flat=True)) == ['bar1', 'bar2']
        queryset = remove_filter(queryset, lookup='username')
        assert list(queryset.values_list('username', flat=True)) == [
            'bar1', 'bar2', 'foo1', 'foo2'
        ]

        queryset = queryset.filter(Q(username='bar1') | Q(username='bar2'))
        assert list(queryset.values_list('username', flat=True)) == ['bar1', 'bar2']
        queryset = remove_filter(queryset, lookup='username')
        assert list(queryset.values_list('username', flat=True)) == [
            'bar1', 'bar2', 'foo1', 'foo2'
        ]

        # Existing model field, but not filtered -> no problem:
        remove_filter(queryset, lookup='email')

        # Try to remove a not existing model field -> error:
        with self.assertRaises(FieldError) as cm:
            remove_filter(queryset, lookup='doesnotexists')
        error_msg = cm.exception.args[0]
        assert error_msg.startswith(
            "Cannot resolve keyword 'doesnotexists' into field. Choices are: "
        )

        # Test with a "empty" queryset:
        empty_queryset1 = User.objects.none()
        assert empty_queryset1.count() == 0
        empty_queryset2 = remove_filter(empty_queryset1, lookup='username')
        assert empty_queryset2.count() == 0
        assert empty_queryset1 is empty_queryset2  # In this case we get the same object back

    def test_remove_model_filter(self):
        admins = Group.objects.create(name='admins')
        customers = Group.objects.create(name='customers')
        User.objects.create(username='Beatrice').groups.add(admins)
        User.objects.create(username='Benedict').groups.add(customers)
        User.objects.create(username='Bob').groups.add(admins)
        User.objects.create(username='Jessica').groups.add(admins)
        User.objects.create(username='John').groups.add(customers)

        # Construct sample query
        queryset = (
            User.objects
            .filter(username__startswith='B')
            .filter(groups__name__in=['admins'])
            .order_by('username').all())
        got = list(queryset.values_list('username', flat=True))
        assert got == ['Beatrice', 'Bob']

        # Remove model-referencing filter
        old_id = id(queryset)
        queryset = remove_model_filter(queryset, Group)
        got = list(queryset.values_list('username', flat=True))
        assert got == ['Beatrice', 'Benedict', 'Bob']
        assert id(queryset) != old_id

        # Not filtering by the model in the first place is no problem
        simple_queryset = User.objects.filter(username__contains='e')
        simple_queryset = remove_model_filter(simple_queryset, Group)
        got = list(simple_queryset.values_list('username', flat=True))
        assert got == ['Beatrice', 'Benedict', 'Jessica']

        # Test with an empty queryset
        empty_queryset1 = User.objects.none()
        assert empty_queryset1.count() == 0
        empty_queryset2 = remove_model_filter(empty_queryset1, Group)
        assert empty_queryset2.count() == 0
        assert empty_queryset1 is empty_queryset2  # In this case we get the same object back



"""
    Use Playwright in Unittest + Fast Django user login
"""
import dataclasses
import os
import warnings

from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test import tag
from playwright.sync_api import Browser, BrowserType, Playwright, sync_playwright


@dataclasses.dataclass
class PlaywrightConfig:
    """
    PlaywrightTestCase config from environment (PWBROWSER, PWHEADLESS, PWSKIP, PWSLOWMO)
    """

    browser_name: str  # Browser to launch: "chromium", "firefox" or "webkit"
    headless: bool  # whether to show the browser during test execution or not
    skip: bool  # whether to skip all Playwright tests or not
    slow_mo: int  # add a delay in microseconds between every interaction

    @classmethod
    def from_env(cls):
        return cls(
            browser_name=os.environ.get('PWBROWSER', 'chromium'),
            headless=bool(int(os.environ.get('PWHEADLESS', '1'))),
            skip=bool(int(os.environ.get('PWSKIP', '0'))),
            slow_mo=int(os.environ.get('PWSLOWMO', '0')),
        )


@tag('playwright')
class PlaywrightTestCase(StaticLiveServerTestCase):
    """
    StaticLiveServerTestCase with helpers for writing frontend tests using Playwright.

    Example usage:

        class MyFrontendTestCase(PlaywrightTestCase):
            def test_foobar(self):
                context = self.browser.new_context(ignore_https_errors=True)
                with context.new_page() as page:
                    page.goto(f'{self.live_server_url}/foo/bar')
                    expect(page).to_have_title('My Site')
    """

    # Optional: Used browser for this test case.
    # If this is set, the environment variable "PWBROWSER" will be ignored.
    # Possible names are: "chromium", "firefox" or "webkit"
    BROWSER_NAME = None

    @classmethod
    def setUpClass(cls):
        """
        Launch Playwright browser init with config from environment.
        """
        super().setUpClass()

        cls.pw_config = PlaywrightConfig.from_env()

        # skip init if "PWSKIP=1" in environment:
        if not cls.pw_config.skip:
            cls.playwright: Playwright = sync_playwright().start()
            cls.browser_name: str = cls.BROWSER_NAME or cls.pw_config.browser_name
            cls.browser_type: BrowserType = getattr(cls.playwright, cls.browser_name)
            cls.browser: Browser = cls.browser_type.launch(
                headless=cls.pw_config.headless,
                slow_mo=cls.pw_config.slow_mo,
            )

            if hasattr(cls, 'setUpTestData'):
                warnings.warn(
                    f'Remove setUpTestData() from {cls.__name__}'
                    ' because it will never be called!'
                )

    @classmethod
    def tearDownClass(cls):
        if not cls.pw_config.skip:
            cls.browser.close()
            cls.playwright.stop()
        super().tearDownClass()

    def setUp(self):
        super().setUp()

        # skip this test if "PWSKIP=1" in environment:
        if self.pw_config.skip:
            self.skipTest('Playwright tests are excluded')



from django.core.exceptions import ValidationError
from django.test import TestCase

from bx_django_utils_tests.test_app.models import ColorFieldTestModel


class ColorFieldTestCase(TestCase):
    def test_model(self):
        instance = ColorFieldTestModel(required_color='#00AAff')
        instance.full_clean()
        assert instance.required_color == '#00AAff'
        assert instance.optional_color is None

        instance = ColorFieldTestModel(required_color='#00XX00')
        msg = "{'required_color': ['Enter a valid seven-character hexadecimal color value.']}"
        with self.assertRaisesMessage(ValidationError, msg):
            instance.full_clean()

        instance = ColorFieldTestModel(required_color='#11223')
        msg = (
            "{'required_color': ['Ensure this value has at least 7 characters (it has 6).',"
            " 'Enter a valid seven-character hexadecimal color value.']}"
        )
        with self.assertRaisesMessage(ValidationError, msg):
            instance.full_clean()

        instance = ColorFieldTestModel(required_color='#1122334')
        msg = (
            "{'required_color': ['Ensure this value has at most 7 characters (it has 8).',"
            " 'Enter a valid seven-character hexadecimal color value.']}"
        )
        with self.assertRaisesMessage(ValidationError, msg):
            instance.full_clean()



import datetime

from django.utils.html import avoid_wrapping
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy


TIMESINCE_CHUNKS = (
    (60 * 60 * 24 * 365, gettext_lazy('%.1f years')),
    (60 * 60 * 24 * 30, gettext_lazy('%.1f months')),
    (60 * 60 * 24 * 7, gettext_lazy('%.1f weeks')),
    (60 * 60 * 24, gettext_lazy('%.1f days')),
    (60 * 60, gettext_lazy('%.1f hours')),
    (60, gettext_lazy('%.1f minutes')),
)


def human_timedelta(t):
    """
    Converts a time duration into a friendly text representation. (`X ms`, `sec`, `minutes` etc.)

    >>> human_timedelta(datetime.timedelta(microseconds=1000))
    '1.0\xa0ms'
    >>> human_timedelta(0.01)
    '10.0\xa0ms'
    >>> human_timedelta(0.9)
    '900.0\xa0ms'
    >>> human_timedelta(datetime.timedelta(seconds=1))
    '1.0\xa0seconds'
    >>> human_timedelta(65.5)
    '1.1\xa0minutes'
    >>> human_timedelta(59 * 60)
    '59.0\xa0minutes'
    >>> human_timedelta(60*60)
    '1.0\xa0hours'
    >>> human_timedelta(1.05*60*60)
    '1.1\xa0hours'
    >>> human_timedelta(datetime.timedelta(hours=24))
    '1.0\xa0days'
    >>> human_timedelta(2.54 * 60 * 60 * 24 * 365)
    '2.5\xa0years'
    >>> human_timedelta('type error')
    Traceback (most recent call last):
        ...
    TypeError: human_timedelta() argument must be timedelta, integer or float)
    """
    if isinstance(t, datetime.timedelta):
        t = t.total_seconds()
    elif not isinstance(t, (int, float)):
        raise TypeError('human_timedelta() argument must be timedelta, integer or float)')

    if abs(t) < 1:
        return avoid_wrapping(_('%.1f ms') % round(t * 1000, 1))
    if abs(t) < 60:
        return avoid_wrapping(_('%.1f seconds') % round(t, 1))

    for seconds, time_string in TIMESINCE_CHUNKS:
        count = t / seconds
        if abs(count) >= 1:
            count = round(count, 1)
            break

    return avoid_wrapping(time_string % count)


human_timedelta.is_safe = True



# Generated by Django 3.2.4 on 2021-09-22 14:59

import bx_django_utils.data_types.gtin.model_fields
import bx_django_utils.data_types.gtin.validators
import bx_django_utils.models.color_field
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ColorFieldTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('required_color', bx_django_utils.models.color_field.ColorModelField(max_length=7)),
                ('optional_color', bx_django_utils.models.color_field.ColorModelField(blank=True, max_length=7, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='GtinFieldTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('default_gtin', bx_django_utils.data_types.gtin.model_fields.GtinModelField(max_length=14, validators=[bx_django_utils.data_types.gtin.validators.GtinValidator(accepted_length=(12, 13, 14))], verbose_name='GTIN')),
                ('all_gtin', bx_django_utils.data_types.gtin.model_fields.GtinModelField(blank=True, max_length=14, validators=[bx_django_utils.data_types.gtin.validators.GtinValidator(accepted_length=(12, 13, 14))], verbose_name='GTIN')),
                ('ean13', bx_django_utils.data_types.gtin.model_fields.GtinModelField(blank=True, max_length=14, null=True, validators=[bx_django_utils.data_types.gtin.validators.GtinValidator(accepted_length=(12, 13, 14))], verbose_name='GTIN')),
            ],
        ),
        migrations.CreateModel(
            name='TimetrackingTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('create_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.create_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.create_dt.verbose_name')),
                ('update_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.update_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.update_dt.verbose_name')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CreateOrUpdateTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('create_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.create_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.create_dt.verbose_name')),
                ('update_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.update_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.update_dt.verbose_name')),
                ('name', models.CharField(max_length=64)),
                ('slug', models.SlugField(max_length=64)),
                ('blank_field', models.CharField(blank=True, max_length=64)),
                ('null_field', models.CharField(blank=True, max_length=64, null=True)),
                ('many2one_rel', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='test_app.timetrackingtestmodel')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from bx_django_utils.test_utils.content_types import ContentTypeCacheFixMixin


def assert_filled_content_type_cache():
    assert ContentType.objects._cache.keys() == {'default'}
    assert len(ContentType.objects._cache['default']) >= 20


class ContentTypesUtilsTestCase(TestCase):
    def test_fill_content_type_cache(self):
        ContentType.objects._cache.clear()

        assert len(ContentType.objects._cache) == 0

        ContentTypeCacheFixMixin._fill_content_type_cache()

        assert_filled_content_type_cache()


class ContentTypeCacheFixMixinTestCase(ContentTypeCacheFixMixin, TestCase):
    """
    Test if the cache filled for all test methods via setUp()
    """

    def test_filled_cache1(self):
        assert_filled_content_type_cache()
        ContentType.objects._cache.clear()

    def test_filled_cache2(self):
        assert_filled_content_type_cache()
        ContentType.objects._cache.clear()



from importlib import import_module

from django.apps import apps
from django.utils.module_loading import module_has_submodule


MODULE_NAME = 'admin_views'


def autodiscover_admin_views():
    found_modules = []
    for app_config in apps.get_app_configs():
        if module_has_submodule(app_config.module, MODULE_NAME):
            module = import_module(f'{app_config.name}.{MODULE_NAME}')
            found_modules.append(module)
    return found_modules



import django
from django.test import TestCase
from model_bakery import baker

from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin
from bx_django_utils.test_utils.users import make_test_user
from bx_django_utils_tests.test_app.models import CreateOrUpdateTestModel


class NotAllSimpleListFilterTestCase(HtmlAssertionMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.superuser = make_test_user(is_superuser=True)

        baker.make(
            CreateOrUpdateTestModel,
            pk=1,
            name='with empty blank field',
            slug='empty',
            blank_field='',
        )
        baker.make(
            CreateOrUpdateTestModel,
            pk=2,
            name='with filled blank field',
            slug='filled',
            blank_field='a value',
        )

    def _get_changelist_items(self, response):
        results = response.context['results']
        items = [items[-1] for items in results]
        return items

    def test_admin(self):
        self.client.force_login(self.superuser)

        ###################################################################################
        # without filter -> default

        response = self.client.get('/admin/test_app/createorupdatetestmodel/')
        if django.VERSION >= (4, 1):
            self.assert_html_parts(
                response,
                parts=(
                    '<summary>By blank_field set</summary>',
                    '<li><a href="?blank_field=all">All</a></li>',
                    '<li class="selected"><a href="?">Yes</a></li>',  # Yes is default!
                    '<li><a href="?blank_field=no">No</a></li>',
                ),
            )
        else:
            self.assert_html_parts(
            response,
            parts=(
                '<h3>By blank_field set</h3>',
                '<li><a href="?blank_field=all" title="All">All</a></li>',
                '<li class="selected"><a href="?" title="Yes">Yes</a></li>',  # Yes is default!
                '<li><a href="?blank_field=no" title="No">No</a></li>',
            ),
        )
        # Only "filled" item?
        items = self._get_changelist_items(response)
        self.assertEqual(
            items,
            [
                '<td class="field-slug">'
                '<a href="/admin/test_app/createorupdatetestmodel/2/change/">filled</a></td>'
            ],
        )

        ###################################################################################
        # only blank fields

        response = self.client.get('/admin/test_app/createorupdatetestmodel/?blank_field=no')
        if django.VERSION >= (4, 1):
            self.assert_html_parts(
                response,
                parts=(
                    '<summary>By blank_field set</summary>',
                    '<li><a href="?blank_field=all">All</a></li>',
                    '<li><a href="?">Yes</a></li>',  # Yes is default!
                    '<li class="selected"><a href="?blank_field=no">No</a></li>',
                ),
            )
        else:
            self.assert_html_parts(
                response,
                parts=(
                    '<h3>By blank_field set</h3>',
                    '<li><a href="?blank_field=all" title="All">All</a></li>',
                    '<li><a href="?" title="Yes">Yes</a></li>',
                    '<li class="selected"><a href="?blank_field=no" title="No">No</a></li>',
                ),
            )
        # Only "empty" item?
        items = self._get_changelist_items(response)
        self.assertEqual(
            items,
            [
                '<td class="field-slug">'
                '<a href="/admin/test_app/createorupdatetestmodel/1/change/'
                '?_changelist_filters=blank_field%3Dno">empty</a></td>'
            ],
        )

        ###################################################################################
        # all entries

        response = self.client.get('/admin/test_app/createorupdatetestmodel/?blank_field=all')
        if django.VERSION >= (4, 1):
            self.assert_html_parts(
                response,
                parts=(
                    '<summary>By blank_field set</summary>',
                    '<li class="selected"><a href="?blank_field=all">All</a></li>',
                    '<li><a href="?">Yes</a></li>',  # Yes is default!
                    '<li><a href="?blank_field=no">No</a></li>',
                ),
            )
        else:
            self.assert_html_parts(
                response,
                parts=(
                    '<h3>By blank_field set</h3>',
                    '<li class="selected"><a href="?blank_field=all" title="All">All</a></li>',
                    '<li><a href="?" title="Yes">Yes</a></li>',
                    '<li><a href="?blank_field=no" title="No">No</a></li>',
                ),
            )
        # Both: "filled" and "empty" items?
        items = self._get_changelist_items(response)
        self.assertEqual(
            items,
            [
                '<td class="field-slug"><a href="/admin/test_app/createorupdatetestmodel/2/change/'
                '?_changelist_filters=blank_field%3Dall">filled</a></td>',
                '<td class="field-slug"><a href="/admin/test_app/createorupdatetestmodel/1/change/'
                '?_changelist_filters=blank_field%3Dall">empty</a></td>',
            ],
        )


class ExistingCountedListFilterTestCase(HtmlAssertionMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.superuser = make_test_user(is_superuser=True)

        baker.make(
            CreateOrUpdateTestModel,
            pk=1,
            name='<foo> / <bar>',
            slug='foo-bar',
            blank_field='a value',
        )
        baker.make(
            CreateOrUpdateTestModel,
            pk=2,
            name='Foo',
            slug='foo',
            blank_field='a value',
        )
        baker.make(
            CreateOrUpdateTestModel,
            pk=3,
            name='Foo',
            slug='foo2',
            blank_field='a value',
        )

    def _get_changelist_items(self, response):
        results = response.context['results']
        items = [items[-1] for items in results]
        return items

    def test_admin(self):
        self.client.force_login(self.superuser)

        response = self.client.get('/admin/test_app/createorupdatetestmodel/')
        if django.VERSION >= (4, 1):
            self.assert_html_parts(
                response,
                parts=(
                    '<summary>By name</summary>',
                    (
                        '<li>'
                        '<a href="?name=%253Cfoo%253E%2B%252F%2B%253Cbar%253E">'
                        '&lt;foo&gt; / &lt;bar&gt; (1)</a>'
                        '</li>'
                    ),
                    '<li><a href="?name=Foo">Foo (2)</a></li>',
                ),
            )
        else:
            self.assert_html_parts(
                response,
                parts=(
                    '<h3>By name</h3>',
                    (
                        '<li>'
                        '<a href="?name=%253Cfoo%253E%2B%252F%2B%253Cbar%253E"'
                        ' title="&lt;foo&gt; / &lt;bar&gt; (1)">'
                        '&lt;foo&gt; / &lt;bar&gt; (1)'
                        '</a></li>'
                    ),
                    '<li><a href="?name=Foo" title="Foo (2)">Foo (2)</a></li>',
                ),
            )
        items = self._get_changelist_items(response)
        self.assertEqual(
            items,
            [
                (
                    '<td class="field-slug">'
                    '<a href="/admin/test_app/createorupdatetestmodel/3/change/">'
                    'foo2</a></td>'
                ),
                (
                    '<td class="field-slug">'
                    '<a href="/admin/test_app/createorupdatetestmodel/2/change/">'
                    'foo</a></td>'
                ),
                (
                    '<td class="field-slug">'
                    '<a href="/admin/test_app/createorupdatetestmodel/1/change/">'
                    'foo-bar</a></td>'
                ),
            ],
        )

        ###################################################################################

        response = self.client.get(
            '/admin/test_app/createorupdatetestmodel/?name=%253Cfoo%253E%2B%252F%2B%253Cbar%253E'
        )
        if django.VERSION >= (4, 1):
            self.assert_html_parts(
                response,
                parts=(
                    '<summary>By name</summary>',
                    (
                        '<li class="selected">'
                        '<a href="?name=%253Cfoo%253E%2B%252F%2B%253Cbar%253E">'
                        '&lt;foo&gt; / &lt;bar&gt; (1)</a>'
                        '</li>'
                    ),
                    '<li><a href="?name=Foo">Foo (2)</a></li>',
                ),
            )
        else:
            self.assert_html_parts(
                response,
                parts=(
                    '<h3>By name</h3>',
                    (
                        '<li class="selected">'
                        '<a href="?name=%253Cfoo%253E%2B%252F%2B%253Cbar%253E"'
                        ' title="&lt;foo&gt; / &lt;bar&gt; (1)">'
                        '&lt;foo&gt; / &lt;bar&gt; (1)'
                        '</a></li>'
                    ),
                    '<li><a href="?name=Foo" title="Foo (2)">Foo (2)</a></li>',
                ),
            )
        items = self._get_changelist_items(response)
        self.assertEqual(
            items,
            [
                '<td class="field-slug">'
                '<a href="/admin/test_app/createorupdatetestmodel/1/change/'
                '?_changelist_filters=name%3D%25253Cfoo%25253E%252B%25252F%252B%25253Cbar%25253E">'
                'foo-bar</a></td>'
            ],
        )



import logging

from bx_py_utils.test_utils.snapshot import assert_snapshot
from django.contrib.auth.models import User
from django.test import TestCase

from bx_django_utils.admin_extra_views.utils import iter_admin_extra_views_urls


class AdminExtraViewsTestCase(TestCase):
    """
    Integrations tests for Admin Extra Views.
    Should be a copy&paste template for bx_django_utils user projects.
    """

    def test_anonymous_access(self):
        checked_urls = 0
        for url in iter_admin_extra_views_urls():
            with self.assertLogs('bx_django_utils', level=logging.WARNING) as logs, self.assertLogs(
                'django.request', level=logging.WARNING
            ):
                response = self.client.get(url)
            logs = '\n'.join(logs.output)
            self.assertIn('did not pass', logs)
            self.assertEqual(response.status_code, 403)
            checked_urls += 1
        self.assertGreaterEqual(checked_urls, 1)

    def test_superuser_access(self):
        superuser = User.objects.create_superuser(username='foobar')
        self.client.force_login(superuser)

        all_urls = []
        for url in iter_admin_extra_views_urls():
            respose = self.client.get(url)
            self.assertEqual(respose.status_code, 200)
            all_urls.append(url)
        self.assertGreaterEqual(len(all_urls), 1)
        assert_snapshot(got=all_urls)



import dataclasses
from typing import Callable, List, Set

from django.contrib.admin import AdminSite
from django.contrib.admin.sites import site as default_site
from django.utils.text import slugify

from bx_django_utils.admin_extra_views.conditions import only_staff_user


_APP_LABELS = set()
_URL_NAMES = set()


@dataclasses.dataclass
class AdminExtraMeta:
    """
    Stores information for pseudo app and pseudo models.
    """

    name: str  # localized display name
    app_label: str = None  # Optional: Will be set from name, if not defined

    # must all be met to display. If not set: only_staff_user() will be added:
    conditions: Set[Callable] = dataclasses.field(default_factory=set)

    # Set via set_url_info():
    url: str = None
    url_name: str = None

    def __post_init__(self):
        if not self.app_label:
            self.app_label = slugify(self.name)

        elif slugify(self.app_label) != self.app_label:
            raise AssertionError(f'App label is not a valid slug: {self.app_label!r}')

        if not self.conditions:
            # Add default condition if nothing set:
            self.conditions.add(only_staff_user)

    def __eq__(self, other):
        return self.app_label == other.app_label

    def __hash__(self):
        return hash(self.app_label)

    def setup_app(self, pseudo_app):
        # Build URL information with the given parent pseudo app:
        self.url_name = f'{pseudo_app.meta.app_label}_{self.app_label}'
        if self.url_name in _URL_NAMES:
            raise AssertionError(f'App label combination {self.url_name!r} is not unique!')
        _URL_NAMES.add(self.url_name)

        self.url = f'{pseudo_app.meta.app_label}/{self.app_label}/'

        # Add parent pseudo app conditions:
        self.conditions = self.conditions.union(pseudo_app.meta.conditions)


@dataclasses.dataclass
class PseudoApp:
    """
    Represents information about a Django App. Instance must be pass to @register_admin_view()
    """

    meta: AdminExtraMeta = None

    # Will be filled by @register_admin_view():
    views: List = dataclasses.field(default_factory=list)

    # The Django Admin site for this extra views (optional):
    admin_site: AdminSite = None

    def __post_init__(self):
        assert isinstance(self.meta, AdminExtraMeta)

        if self.meta.app_label in _APP_LABELS:
            raise AssertionError(
                f'PseudoApp must be have a unique label! Current label is: {self.meta.app_label!r}'
            )
        _APP_LABELS.add(self.meta.app_label)

        if not self.admin_site:
            self.admin_site = default_site

    def __eq__(self, other):
        return self.meta.app_label == other.meta.app_label

    def __hash__(self):
        return hash(self.meta.app_label)



import pprint

from django.core.serializers.json import DjangoJSONEncoder


def pformat(value):
    """
    Better `pretty-print-format` using `DjangoJSONEncoder` with fallback to `pprint.pformat()`

    Try to use JSON fist, because it nicer than pprint.pformat() ;)
    Use DjangoJSONEncoder, because it can encode additional types like:
        date/time, decimal types, UUIDs
    """
    try:
        value = DjangoJSONEncoder(indent=4, sort_keys=True, ensure_ascii=False).encode(value)
    except TypeError:
        # Fallback if values are not serializable with JSON:
        value = pprint.pformat(value, width=120)

    return value



from django.contrib.auth.models import User
from django.test import TestCase
from model_bakery import baker

from bx_django_utils.admin_utils.admin_urls import (
    admin_change_url,
    admin_changelist_url,
    admin_delete_url,
    admin_history_url,
    admin_model_url,
)


class AdminUrlsSimpleTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = baker.make(User, id=1)

    def test_admin_model_url(self):
        # defaults to "change":
        self.assertEqual(admin_model_url(model_or_instance=self.user), '/admin/auth/user/1/change/')

        # defaults to "changelist":
        self.assertEqual(admin_model_url(model_or_instance=User), '/admin/auth/user/')

        # All action variants for a instance:
        self.assertEqual(
            admin_model_url(model_or_instance=self.user, action='change'),
            '/admin/auth/user/1/change/',
        )
        self.assertEqual(
            admin_model_url(model_or_instance=self.user, action='history'),
            '/admin/auth/user/1/history/',
        )
        self.assertEqual(
            admin_model_url(model_or_instance=self.user, action='delete'),
            '/admin/auth/user/1/delete/',
        )

        # All action variants for a model class:
        self.assertEqual(
            admin_model_url(model_or_instance=User, action='changelist'), '/admin/auth/user/'
        )
        self.assertEqual(
            admin_model_url(model_or_instance=User, action='add'), '/admin/auth/user/add/'
        )

        # Optional parameters:
        url = admin_model_url(model_or_instance=User, params=dict(is_staff__exact=1))
        self.assertEqual(url, '/admin/auth/user/?is_staff__exact=1')

        with self.assertRaisesMessage(
            AssertionError, "Action: 'Bam!' is not one of changelist, add, history, delete, change"
        ):
            admin_model_url(model_or_instance=User, action='Bam!')

        with self.assertRaisesMessage(AssertionError, 'is no model instance'):
            admin_model_url(model_or_instance=User, action='change')

    def test_admin_change_url(self):
        self.assertEqual(admin_change_url(self.user), '/admin/auth/user/1/change/')

    def test_admin_history_url(self):
        self.assertEqual(admin_history_url(self.user), '/admin/auth/user/1/history/')

    def test_admin_delete_url(self):
        self.assertEqual(admin_delete_url(self.user), '/admin/auth/user/1/delete/')

    def test_admin_changelist_url(self):
        self.assertEqual(admin_changelist_url(self.user), '/admin/auth/user/')
        self.assertEqual(admin_changelist_url(User), '/admin/auth/user/')



"""
    Just a DEMO for bx_django_utils.admin_extra_views
"""
from django.contrib.auth.models import User
from django.http import HttpResponse
from django.urls import reverse
from django.views.generic.base import TemplateView, View

from bx_django_utils.admin_extra_views.base_view import AdminExtraViewMixin
from bx_django_utils.admin_extra_views.datatypes import AdminExtraMeta, PseudoApp
from bx_django_utils.admin_extra_views.registry import register_admin_view


def only_john_can_access(request):
    user: User = request.user
    if user.is_superuser:
        return True
    if user.username == 'john':
        return True
    else:
        return False


pseudo_app1 = PseudoApp(meta=AdminExtraMeta(name='Pseudo App 1'))
pseudo_app2 = PseudoApp(meta=AdminExtraMeta(name='Pseudo App 2'))


@register_admin_view(pseudo_app=pseudo_app1)
class DemoView1(AdminExtraViewMixin, TemplateView):
    meta = AdminExtraMeta(name='Demo View 1')
    template_name = 'admin_extra_views/demo1.html'

    def get_context_data(self, **context):
        context = super().get_context_data(**context)
        context['content'] = 'Just the demo view 1'
        return context


@register_admin_view(pseudo_app=pseudo_app1)
class DemoView2(AdminExtraViewMixin, View):
    meta = AdminExtraMeta(name='Demo View 2')

    def get(self, request):
        return HttpResponse('Just the demo view 2')


@register_admin_view(pseudo_app=pseudo_app2)
class DemoView3(AdminExtraViewMixin, View):
    meta = AdminExtraMeta(
        name='Demo View 3',
        conditions={only_john_can_access},  # Note: pseudo_app2 conditions will be added!
    )

    def get(self, request):
        url = reverse(self.meta.url_name)
        return HttpResponse(f'Just the demo view at {url!r}')



from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from bx_django_utils.templatetags.humanize_time import human_duration


class TimetrackingBaseModel(models.Model):
    """
    Abstract base model that will automaticly set create/update Datetimes.

    Using a DateTimeField with auto_now/auto_now_add is not a good idea.
    Because you can't easy disable these auto behaviour.

    Here you can pass update_dt=False to save() method.
    """
    create_dt = models.DateTimeField(
        blank=True, null=True, editable=False,
        verbose_name=_('ModelTimetrackingMixin.create_dt.verbose_name'),
        help_text=_('ModelTimetrackingMixin.create_dt.help_text')
    )
    update_dt = models.DateTimeField(
        blank=True, null=True, editable=False,
        verbose_name=_('ModelTimetrackingMixin.update_dt.verbose_name'),
        help_text=_('ModelTimetrackingMixin.update_dt.help_text')
    )

    def human_create_dt(self):
        return human_duration(self.create_dt)
    human_create_dt.short_description = _('ModelTimetrackingMixin.create_dt.verbose_name')
    human_create_dt.admin_order_field = 'create_dt'

    def human_update_dt(self):
        return human_duration(self.update_dt)
    human_update_dt.short_description = _('ModelTimetrackingMixin.update_dt.verbose_name')
    human_update_dt.admin_order_field = 'update_dt'

    def save(self, update_dt=True, **kwargs):
        if update_dt:
            if 'update_fields' in kwargs:
                update_fields = list(kwargs['update_fields'])
            else:
                update_fields = None

            self.update_dt = timezone.now()
            if update_fields is not None:
                assert 'update_dt' not in update_fields
                update_fields.append('update_dt')

            if self.create_dt is None:
                self.create_dt = self.update_dt
                if update_fields is not None:
                    assert 'create_dt' not in update_fields
                    update_fields.append('create_dt')

            if update_fields is not None:
                kwargs['update_fields'] = update_fields

        super().save(**kwargs)

    class Meta:
        abstract = True



import logging

from django.core.exceptions import PermissionDenied

from bx_django_utils.admin_extra_views.datatypes import AdminExtraMeta


logger = logging.getLogger(__name__)


class AdminExtraViewMixin:
    meta: AdminExtraMeta = None  # Must be set by view

    def dispatch(self, request, *args, **kwargs):
        for condition_func in self.meta.conditions:
            if not condition_func(request):
                logger.warning(
                    'User (pk:%r) did not pass %r for %r',
                    request.user.pk,
                    condition_func.__name__,
                    self.__class__.__name__,
                )
                raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)



from django.apps import AppConfig

from bx_django_utils.admin_extra_views.autodiscover import autodiscover_admin_views


class AdminExtraViewsAppConfig(AppConfig):
    """
    App config to auto discover all extra views.
    """

    name = 'bx_django_utils.admin_extra_views'
    verbose_name = "Admin Extra views"

    def ready(self):
        autodiscover_admin_views()

        # Register system checks:
        from bx_django_utils.admin_extra_views.checks import admin_extra_views_check  # noqa



import dataclasses
from uuid import UUID

from django.core.cache import cache


@dataclasses.dataclass
class CachedDataclassBase:
    """
    A Base dataclass that can be easy store/restore to Django cache.
    """

    # Sady, it's not possible to mix non-default and default arguments with inheritance.
    # So the child class must define this:
    #
    # uuid: UUID = dataclasses.field(default_factory=uuid4)
    #

    def __post_init__(self):
        # Check if child class set a UUID:
        assert hasattr(self, 'uuid'), 'Child class must define a "uuid" value!'
        assert isinstance(self.uuid, UUID), f'A UUID is needed. Found: {type(self.uuid).__name__}'

    @property
    def cache_timeout(self):
        """
        Timeout used in cache.set() - To be overwritten, if needed.
        """
        return 1 * 60 * 60

    @classmethod
    def generate_cache_key(cls, uuid):
        prefix = cls.__name__
        cache_key = f'{prefix}-{uuid}'
        return cache_key

    def store2cache(self):
        """
        Save the current data to Django's cache.
        """
        cache_key = self.generate_cache_key(self.uuid)
        data_dict = dataclasses.asdict(self)
        cache.set(cache_key, data_dict, timeout=self.cache_timeout)

    @classmethod
    def get_from_cache(cls, uuid):
        """
        Restore data from Django's cache, if available.
        """
        cache_key = cls.generate_cache_key(uuid=uuid)
        if data_dict := cache.get(cache_key):
            return cls(**data_dict)

    def delete_cache_entry(self):
        """
        Remove data in Django's cache.
        """
        cache_key = self.generate_cache_key(self.uuid)
        return cache.delete(cache_key)



from django import forms
from stdnum.ean import compact

from bx_django_utils.data_types.gtin.constants import DEFAULT_ACCEPT_LENGTH
from bx_django_utils.data_types.gtin.validators import GtinValidator


class GtinFormField(forms.CharField):
    """
    Form field with GTIN validator.
    """

    def __init__(self, *args, accepted_length=DEFAULT_ACCEPT_LENGTH, **kwargs):
        kwargs['min_length'] = min(accepted_length)
        kwargs['max_length'] = min(accepted_length) + 3  # +spaces, e.g.: '6 92771 98116 1'
        kwargs['validators'] = [GtinValidator(accepted_length=accepted_length)]
        super().__init__(*args, **kwargs)

    def clean(self, value):
        value = super().clean(value)
        if value:
            value = compact(value)  # Convert to the minimal representation
        return value

    def prepare_value(self, value):
        if value:
            value = compact(value)  # Convert to the minimal representation
        return value



import datetime

from django.test import SimpleTestCase
from django.utils import translation

from bx_django_utils.humanize.time import human_timedelta


class HumanizeTimeTestCase(SimpleTestCase):
    def test_basic(self):
        with translation.override('en'):
            assert human_timedelta(datetime.timedelta(microseconds=2500)) == '2.5\xa0ms'
            assert human_timedelta(datetime.timedelta(microseconds=-5250)) == '-5.2\xa0ms'

            assert human_timedelta(2.5) == '2.5\xa0seconds'
            assert human_timedelta(-10.3) == '-10.3\xa0seconds'

            assert human_timedelta(datetime.timedelta(days=5 * 365)) == '5.0\xa0years'
            assert human_timedelta(-datetime.timedelta(days=5 * 365)) == '-5.0\xa0years'

        with translation.override('de'):
            assert human_timedelta(datetime.timedelta(microseconds=2500)) == '2.5\xa0ms'
            assert human_timedelta(2.5) == '2.5\xa0Sekunden'
            assert human_timedelta(datetime.timedelta(days=3, hours=12)) == '3.5\xa0Tage'
            assert human_timedelta(datetime.timedelta(days=5 * 365)) == '5.0\xa0Jahre'




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bx_django_utils',
 'bx_django_utils.admin_extra_views',
 'bx_django_utils.admin_extra_views.management',
 'bx_django_utils.admin_extra_views.management.commands',
 'bx_django_utils.admin_extra_views.tests',
 'bx_django_utils.admin_utils',
 'bx_django_utils.approve_workflow',
 'bx_django_utils.data_types',
 'bx_django_utils.data_types.gtin',
 'bx_django_utils.dbperf',
 'bx_django_utils.humanize',
 'bx_django_utils.models',
 'bx_django_utils.templatetags',
 'bx_django_utils.test_utils',
 'bx_django_utils.user_timezone',
 'bx_django_utils.user_timezone.templatetags',
 'bx_django_utils.view_utils',
 'bx_django_utils_tests',
 'bx_django_utils_tests.approve_workflow_test_app',
 'bx_django_utils_tests.approve_workflow_test_app.migrations',
 'bx_django_utils_tests.test_app',
 'bx_django_utils_tests.test_app.management',
 'bx_django_utils_tests.test_app.management.commands',
 'bx_django_utils_tests.test_app.migrations',
 'bx_django_utils_tests.test_app.tests',
 'bx_django_utils_tests.test_app.tests.fixtures',
 'bx_django_utils_tests.test_app.views',
 'bx_django_utils_tests.test_project',
 'bx_django_utils_tests.tests',
 'bx_django_utils_tests.tests.admin_utils',
 'bx_django_utils_tests.tests.test_user_timezone']

package_data = \
{'': ['*'],
 'bx_django_utils': ['locale/de/LC_MESSAGES/*', 'locale/en/LC_MESSAGES/*'],
 'bx_django_utils.approve_workflow': ['locale/de/LC_MESSAGES/*',
                                      'locale/en/LC_MESSAGES/*',
                                      'templates/approve_workflow/*'],
 'bx_django_utils.user_timezone': ['static/*', 'templates/user_timezone/*'],
 'bx_django_utils_tests.test_app': ['templates/admin_extra_views/*',
                                    'templates/dynamic_view_menu/*'],
 'bx_django_utils_tests.test_project': ['templates/admin/*']}

install_requires = \
['bx_py_utils', 'django>=3.2,<4.2', 'python-stdnum']

entry_points = \
{'console_scripts': ['publish = '
                     'bx_django_utils_tests.test_project.publish:publish']}

setup_kwargs = {
    'name': 'bx-django-utils',
    'version': '39',
    'description': 'Various Django utility functions',
    'long_description': '# Boxine - bx_django_utils\n\nVarious Django utility functions\n\n[![unittests](https://github.com/boxine/bx_django_utils/actions/workflows/pythonapp.yml/badge.svg?branch=master)](https://github.com/boxine/bx_django_utils/actions/workflows/pythonapp.yml) [![Coverage Status on codecov.io](https://codecov.io/gh/boxine/bx_django_utils/branch/master/graph/badge.svg)](https://codecov.io/gh/boxine/bx_django_utils)\n\n[![bx_django_utils @ PyPi](https://img.shields.io/pypi/v/bx_django_utils?label=bx_django_utils%20%40%20PyPi)](https://pypi.org/project/bx_django_utils/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/bx_django_utils)](https://gitlab.com/boxine/bx_django_utils/-/blob/main/pyproject.toml)\n[![License MIT](https://img.shields.io/pypi/l/bx_django_utils)](https://gitlab.com/boxine/bx_django_utils/-/blob/main/LICENSE)\n\n\n## Quickstart\n\n```bash\npip install bx_django_utils\n```\n\n\n## Existing stuff\n\nHere only a simple list about existing utilities.\nPlease take a look into the sources and tests for deeper informations.\n\n[comment]: <> (✂✂✂ auto generated start ✂✂✂)\n\n### bx_django_utils.admin_extra_views\n\nDjango Admin extra views: https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/README.md\n\n\n#### bx_django_utils.admin_extra_views.admin_config\n\nActivate "ExtraViewAdminSite" by set this as default admin site\n\n* [`CustomAdminConfig()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/admin_config.py#L7-L12) - Change Django Admin Site to ExtraViewAdminSite for the extra views.\n\n#### bx_django_utils.admin_extra_views.apps\n\n* [`AdminExtraViewsAppConfig()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/apps.py#L6-L18) - App config to auto discover all extra views.\n\n#### bx_django_utils.admin_extra_views.conditions\n\n* [`only_staff_user()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/conditions.py#L4-L15) - Pass only active staff users. The default condition for all admin extra views.\n\n#### bx_django_utils.admin_extra_views.datatypes\n\n* [`AdminExtraMeta()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/datatypes.py#L15-L58) - Stores information for pseudo app and pseudo models.\n* [`PseudoApp()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/datatypes.py#L61-L91) - Represents information about a Django App. Instance must be pass to @register_admin_view()\n\n###### bx_django_utils.admin_extra_views.management.commands.admin_extra_views\n\n* [`Command()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/management/commands/admin_extra_views.py#L7-L26) - Manage command "admin_extra_views": Info about registered admin extra views\n\n#### bx_django_utils.admin_extra_views.registry\n\n* [`AdminExtraViewRegistry()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/registry.py#L11-L97) - Hold all information about all admin extra views to expand urls and admin app list.\n* [`register_admin_view()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/registry.py#L103-L116) - Decorator to add a normal view as pseudo App/Model to the admin.\n\n#### bx_django_utils.admin_extra_views.site\n\n* [`ExtraViewAdminSite()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/site.py#L6-L13) - An AdminSite object encapsulates an instance of the Django admin application, ready\n\n##### bx_django_utils.admin_extra_views.tests.test_admin_extra_views\n\n* [`AdminExtraViewsTestCase()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/tests/test_admin_extra_views.py#L10-L39) - Integrations tests for Admin Extra Views.\n\n#### bx_django_utils.admin_extra_views.utils\n\n* [`iter_admin_extra_views_urls()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/utils.py#L20-L26) - Iterate over all registered admin extra view urls.\n* [`reverse_admin_extra_view()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/utils.py#L10-L17) - Get the URL of a Admin Extra View, e.g.: url=reverse_admin_extra_view(YouAdminExtraView)\n\n#### bx_django_utils.admin_extra_views.views\n\n* [`Redirect2AdminExtraView()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_extra_views/views.py#L7-L23) - Redirect to a Admin Extra Views.\n\n#### bx_django_utils.admin_utils.admin_urls\n\nHelpers to build Admin URLs\n\n* [`admin_change_url()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_utils/admin_urls.py#L62-L78) - Shortcut to generate Django admin "change" url for a model instance.\n* [`admin_changelist_url()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_utils/admin_urls.py#L119-L134) - Shortcut to generate Django admin "changelist" url for a model or instance.\n* [`admin_delete_url()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_utils/admin_urls.py#L100-L116) - Shortcut to generate Django admin "delete" url for a model instance.\n* [`admin_history_url()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_utils/admin_urls.py#L81-L97) - Shortcut to generate Django admin "history" url for a model instance.\n* [`admin_model_url()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_utils/admin_urls.py#L15-L59) - Build Admin change, add, changelist, etc. links with optional filter parameters.\n\n#### bx_django_utils.admin_utils.filters\n\n* [`ExistingCountedListFilter()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_utils/filters.py#L26-L62) - Advanced SimpleListFilter that list only existing filter values with counts.\n* [`NotAllSimpleListFilter()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/admin_utils/filters.py#L8-L23) - Similar to SimpleListFilter, but don\'t add "All" choice.\n\n### bx_django_utils.approve_workflow\n\nBase model/admin/form classes to implement a model with draft/approve versions workflow\n\n\n#### bx_django_utils.approve_workflow.admin\n\n* [`BaseApproveModelAdmin()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/approve_workflow/admin.py#L15-L107) - Base admin class for a draft/approve Model\n\n#### bx_django_utils.approve_workflow.forms\n\n* [`PublishAdminForm()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/approve_workflow/forms.py#L7-L46) - Activate models REQUIRED_FIELDS_PUBLIC on approve\n\n#### bx_django_utils.approve_workflow.models\n\n* [`BaseApproveModel()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/approve_workflow/models.py#L14-L202) - Base model class for approve models *and* this relation models.\n* [`BaseApproveWorkflowModel()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/approve_workflow/models.py#L205-L255) - Base model for approve workflow models.\n\n### bx_django_utils.cached_dataclasses\n\n* [`CachedDataclassBase()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/cached_dataclasses.py#L7-L59) - A Base dataclass that can be easy store/restore to Django cache.\n\n#### bx_django_utils.data_types.gtin\n\nModelField, FormField and validators for GTIN/UPC/EAN numbers\n\n\n##### bx_django_utils.data_types.gtin.form_fields\n\n* [`GtinFormField()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/data_types/gtin/form_fields.py#L8-L28) - Form field with GTIN validator.\n\n##### bx_django_utils.data_types.gtin.model_fields\n\n* [`GtinModelField()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/data_types/gtin/model_fields.py#L10-L33) - GTIN model field\n\n##### bx_django_utils.data_types.gtin.validators\n\n* [`GtinValidator()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/data_types/gtin/validators.py#L33-L52) - Validate GTIN number\n* [`validate_gtin()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/data_types/gtin/validators.py#L12-L30) - It\'s the same as stdnum.ean.validate() but also accept ISBN-10\n\n#### bx_django_utils.dbperf.cursor\n\n* [`RecordingCursorWrapper()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/dbperf/cursor.py#L17-L136) - An implementation of django.db.backends.utils.CursorWrapper.\n\n#### bx_django_utils.dbperf.query_recorder\n\n* [`SQLQueryRecorder()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/dbperf/query_recorder.py#L95-L176) - A context manager that allows recording SQL queries executed during its lifetime.\n\n### bx_django_utils.filename\n\n* [`clean_filename()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/filename.py#L34-L64) - Convert filename to ASCII only via slugify.\n* [`filename2human_name()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/filename.py#L7-L31) - Convert filename to a capitalized name.\n\n#### bx_django_utils.humanize.pformat\n\n* [`pformat()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/humanize/pformat.py#L6-L20) - Better `pretty-print-format` using `DjangoJSONEncoder` with fallback to `pprint.pformat()`\n\n#### bx_django_utils.humanize.time\n\n* [`human_timedelta()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/humanize/time.py#L18-L63) - Converts a time duration into a friendly text representation. (`X ms`, `sec`, `minutes` etc.)\n\n### bx_django_utils.json_utils\n\n* [`make_json_serializable()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/json_utils.py#L20-L37) - Convert value to a JSON serializable value, with convert callback for special objects.\n* [`to_json()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/json_utils.py#L40-L56) - Convert value to JSON via make_json_serializable() and DjangoJSONEncoder()\n\n#### bx_django_utils.models.color_field\n\n* [`ColorModelField()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/color_field.py#L14-L29) - Hex color model field, e.g.: "#0055ff" (It\'s not a html color picker widget)\n* [`HexColorValidator()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/color_field.py#L6-L11) - Hex color validator (seven-character hexadecimal notation, e.g.: "#0055ff")\n\n#### bx_django_utils.models.manipulate\n\nUtilities to manipulate objects in database via models:\n\n* [`CreateOrUpdateResult()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/manipulate.py#L39-L63) - Result object returned by create_or_update2() with all information about create/save a model.\n* [`InvalidStoreBehavior()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/manipulate.py#L19-L23) - Exception used in create_or_update() if "store_behavior" contains not existing field names.\n* [`create()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/manipulate.py#L26-L36) - Create a new model instance with optional validate before create.\n* [`create_or_update()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/manipulate.py#L197-L217) - Create a new model instance or update a existing one. Deprecated! Use: create_or_update2()\n* [`create_or_update2()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/manipulate.py#L66-L194) - Create a new model instance or update a existing one and returns CreateOrUpdateResult instance\n\n#### bx_django_utils.models.queryset_utils\n\n* [`remove_filter()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/queryset_utils.py#L8-L33) - Remove an applied .filter() from a QuerySet\n* [`remove_model_filter()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/queryset_utils.py#L36-L58) - Remove an applied .filter() from a QuerySet if it contains references to the specified model\n\n#### bx_django_utils.models.timetracking\n\n* [`TimetrackingBaseModel()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/models/timetracking.py#L8-L62) - Abstract base model that will automaticly set create/update Datetimes.\n\n### bx_django_utils.stacktrace\n\n* [`StackTrace()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/stacktrace.py#L21-L22) - Built-in mutable sequence.\n* [`StacktraceAfter()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/stacktrace.py#L83-L111) - Generate a stack trace after a package was visited.\n* [`get_stacktrace()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/stacktrace.py#L63-L80) - Returns a StackTrace object, which is a list of FrameInfo objects.\n\n#### bx_django_utils.templatetags.humanize_time\n\n* [`human_duration()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/templatetags/humanize_time.py#L15-L45) - Verbose time since template tag, e.g.: `<span title="Jan. 1, 2000, noon">2.0 seconds</span>`\n\n### bx_django_utils.test_utils\n\nUtilities / helper for writing tests.\n\n\n#### bx_django_utils.test_utils.assert_queries\n\n* [`AssertQueries()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/assert_queries.py#L34-L281) - Assert executed database queries: Check table names, duplicate/similar Queries.\n\n#### bx_django_utils.test_utils.cache\n\n* [`ClearCacheMixin()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/cache.py#L100-L111) - TestCase mixin to clear the Django cache in setUp/tearDown\n* [`MockCache()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/cache.py#L55-L97) - Mock Django cache backend, so it\'s easy to check/manipulate the cache content\n\n#### bx_django_utils.test_utils.content_types\n\n* [`ContentTypeCacheFixMixin()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/content_types.py#L7-L35) - TestCase mixin to fill the ContentType cache to avoid flaky database queries.\n\n#### bx_django_utils.test_utils.datetime\n\n* [`MockDatetimeGenerator()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/datetime.py#L4-L50) - Mock django `timezone.now()` with generic time stamps in tests.\n\n#### bx_django_utils.test_utils.fixtures\n\nUtilities to manage text fixtures in JSON files.\n\n* [`BaseFixtures()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/fixtures.py#L28-L61) - Base class for JSON dump fixtures.\n* [`FixturesRegistry()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/fixtures.py#L98-L128) - Registry to collect a list of all existing fixture classes.\n* [`RenewAllFixturesBaseCommand()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/fixtures.py#L160-L210) - A base Django manage command to renew all existing fixture JSON dump files\n* [`SerializerFixtures()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/fixtures.py#L64-L95) - Helper to store/restore model instances serialized into a JSON file.\n* [`autodiscover()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/fixtures.py#L134-L157) - Register all fixtures by import all **/fixtures/**/*.py files\n\n#### bx_django_utils.test_utils.forms\n\n* [`AssertFormFields()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/forms.py#L13-L86) - Helper to check the existing form fields.\n\n#### bx_django_utils.test_utils.html_assertion\n\n* [`HtmlAssertionMixin()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/html_assertion.py#L38-L140) - Unittest mixin class with useful assertments around Django test client tests\n* [`assert_html_response_snapshot()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/html_assertion.py#L9-L35) - Assert a HttpResponse via snapshot file using assert_html_snapshot() from bx_py_utils.\n\n#### bx_django_utils.test_utils.model_clean_assert\n\n* [`AssertModelCleanCalled()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/model_clean_assert.py#L35-L86) - Context manager for assert that full_clean() was called for every model instance.\n* [`CleanMock()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/model_clean_assert.py#L6-L32) - Track if full_clean() was called.\n\n#### bx_django_utils.test_utils.playwright\n\nUse Playwright in Unittest + Fast Django user login\n\n* [`PlaywrightConfig()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/playwright.py#L13-L31) - PlaywrightTestCase config from environment (PWBROWSER, PWHEADLESS, PWSKIP, PWSLOWMO)\n* [`PlaywrightTestCase()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/playwright.py#L34-L91) - StaticLiveServerTestCase with helpers for writing frontend tests using Playwright.\n\n#### bx_django_utils.test_utils.users\n\n* [`assert_permissions()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/users.py#L31-L46) - Check user permissions.\n* [`assert_user_properties()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/users.py#L149-L171) - Check a user instance with all properties and password (optional)\n* [`filter_permission_names()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/users.py#L10-L28) - Generate a Permission model query filtered by names, e.g.: [\'<app_label>.<codename>\', ...]\n* [`make_max_test_user()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/users.py#L112-L146) - Create a test user with all permissions *except* the {exclude_permissions} ones.\n* [`make_minimal_test_user()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/users.py#L85-L109) - Create a test user and set given permissions.\n* [`make_test_user()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/test_utils/users.py#L49-L82) - Create a test user and set given permissions.\n\n### bx_django_utils.user_timezone\n\nAutomatic local user timezone: https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/user_timezone/README.md\n\n\n#### bx_django_utils.user_timezone.apps\n\n* [`UserTimezoneAppConfig()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/user_timezone/apps.py#L4-L14) - Django app to set the user local time zone.\n\n#### bx_django_utils.user_timezone.humanize\n\n* [`human_timezone_datetime()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/user_timezone/humanize.py#L7-L21) - Render a datetime with timezone information.\n\n#### bx_django_utils.user_timezone.middleware\n\n* [`UserTimezoneMiddleware()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/user_timezone/middleware.py#L10-L31) - Activate Timezone by "UserTimeZone" cookie\n\n##### bx_django_utils.user_timezone.templatetags.user_timezone\n\n* [`humane_timezone_dt()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/user_timezone/templatetags/user_timezone.py#L9-L14) - Template filter to render a datetime with timezone information.\n\n#### bx_django_utils.view_utils.dynamic_menu_urls\n\n* [`DynamicViewMenu()`](https://github.com/boxine/bx_django_utils/blob/master/bx_django_utils/view_utils/dynamic_menu_urls.py#L4-L45) - Simple storage for store information about views/urls to build a menu.\n\n[comment]: <> (✂✂✂ auto generated end ✂✂✂)\n\n## developing\n\nTo start developing e.g.:\n\n```bash\n~$ git clone https://github.com/boxine/bx_django_utils.git\n~$ cd bx_django_utils\n~/bx_django_utils$ make\nhelp                 List all commands\ninstall-poetry       install poetry\ninstall              install via poetry\nupdate               Update the dependencies as according to the pyproject.toml file\nlint                 Run code formatters and linter\nfix-code-style       Fix code formatting\ntox-listenvs         List all tox test environments\ntox                  Run unittests via tox with all environments\ntest                 Run unittests\npublish              Release new version to PyPi\ndocker-test          Run tests in docker\nmakemessages         Make and compile locales message files\nstart-dev-server     Start Django dev. server with the test project\nclean                Remove created files from the test project (e.g.: SQlite, static files)\nplaywright-install   Install test browser for Playwright tests\nplaywright-inspector Run Playwright inspector\nplaywright-tests     Run only the Playwright tests\n```\n\nYou can start the test project with the Django developing server, e.g.:\n```bash\n~/bx_django_utils$ make start-dev-server\n```\nThis is a own manage command, that will create migrations files from our test app, migrate, collectstatic and create a super user if no user exists ;)\n\nIf you like to start from stretch, just delete related test project files with:\n```bash\n~/bx_django_utils$ make clean\n```\n...and start the test server again ;)\n\n\n## License\n\n[MIT](LICENSE). Patches welcome!\n\n\n## About us\n\nWe’ve been rethinking the listening experience for kids and have created an ecosystem where haptic and listening experience are combined via smart technology - the Toniebox.\n\nWe are constantly looking for engineers to join our team in different areas. If you’d be interested in contributing to our platform, have a look at: https://tonies.com/jobs/\n\n\n## Links\n\n* https://pypi.org/project/bx-django-utils/\n',
    'author': 'Jens Diemer',
    'author_email': 'jens.diemer@boxine.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/boxine/bx_django_utils/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0.0',
}


setup(**setup_kwargs)



from django.contrib.auth.models import User


def only_staff_user(request):
    """
    Pass only active staff users. The default condition for all admin extra views.
    """
    user: User = request.user
    if not user.is_authenticated:
        return False
    if not user.is_active:
        return False
    if not user.is_staff:
        return False
    return True



"""
    Helpers to build Admin URLs
"""
from typing import Optional, Union
from urllib.parse import urlencode

from django.contrib.admin.utils import quote
from django.db import models
from django.urls import reverse


ADMIN_LINK_TYPES = ('changelist', 'add', 'history', 'delete', 'change')


def admin_model_url(
    *,
    model_or_instance: Union[models.Model, type[models.Model]],
    action: Optional[str] = None,  # Default is 'changelist' and 'change'
    admin_prefix: str = 'admin',
    current_app: Optional[str] = None,
    params: Optional[dict] = None,
):
    """
    Build Admin change, add, changelist, etc. links with optional filter parameters.

    'action' defaults to 'changelist' or 'change', depending on model or instance is given.
    """
    if action is None:
        if isinstance(model_or_instance, models.Model):
            action = 'change'
        else:
            action = 'changelist'

    assert (
        action in ADMIN_LINK_TYPES
    ), f'Action: {action!r} is not one of {", ".join(ADMIN_LINK_TYPES)}'

    opts = model_or_instance._meta
    app_label = opts.app_label
    model_name = opts.model_name

    if action in ('history', 'delete', 'change'):
        assert isinstance(
            model_or_instance, models.Model
        ), f'{model_or_instance!r} is no model instance!'
        pk = model_or_instance.pk
        args = (quote(pk),)
    else:
        args = None

    url = reverse(
        f'{admin_prefix}:{app_label}_{model_name}_{action}', current_app=current_app, args=args
    )

    if params:
        params = urlencode(params)
        url = f'{url}?{params}'

    return url


def admin_change_url(
    instance: models.Model,
    admin_prefix: str = 'admin',
    current_app: Optional[str] = None,
    params: Optional[dict] = None,
):
    """
    Shortcut to generate Django admin "change" url for a model instance.
    """
    assert isinstance(instance, models.Model), f'No model instance given: {instance}'
    return admin_model_url(
        model_or_instance=instance,
        action='change',
        admin_prefix=admin_prefix,
        current_app=current_app,
        params=params,
    )


def admin_history_url(
    instance: models.Model,
    admin_prefix: str = 'admin',
    current_app: Optional[str] = None,
    params: Optional[dict] = None,
):
    """
    Shortcut to generate Django admin "history" url for a model instance.
    """
    assert isinstance(instance, models.Model), f'No model instance given: {instance}'
    return admin_model_url(
        model_or_instance=instance,
        action='history',
        admin_prefix=admin_prefix,
        current_app=current_app,
        params=params,
    )


def admin_delete_url(
    instance: models.Model,
    admin_prefix: str = 'admin',
    current_app: Optional[str] = None,
    params: Optional[dict] = None,
):
    """
    Shortcut to generate Django admin "delete" url for a model instance.
    """
    assert isinstance(instance, models.Model), f'No model instance given: {instance}'
    return admin_model_url(
        model_or_instance=instance,
        action='delete',
        admin_prefix=admin_prefix,
        current_app=current_app,
        params=params,
    )


def admin_changelist_url(
    model_or_instance: Union[models.Model, type[models.Model]],
    admin_prefix: str = 'admin',
    current_app: Optional[str] = None,
    params: Optional[dict] = None,
):
    """
    Shortcut to generate Django admin "changelist" url for a model or instance.
    """
    return admin_model_url(
        model_or_instance=model_or_instance,
        action='changelist',
        admin_prefix=admin_prefix,
        current_app=current_app,
        params=params,
    )



from django.contrib import admin

from bx_django_utils.admin_extra_views.registry import extra_view_registry


class ExtraViewAdminSite(admin.AdminSite):
    def get_app_list(self, request):
        app_list = super().get_app_list(request)

        extra_app_list = extra_view_registry.get_app_list(request)
        app_list.extend(extra_app_list)

        return app_list



from bx_py_utils.test_utils.assertion import assert_equal
from django.test import TestCase

from bx_django_utils.stacktrace import get_stacktrace, iter_frameinfo


def foo():
    exclude_modules = (  # keep the test stacktrace minimal
        'unittest',
        'django.test',
        'runpy',
    )
    return get_stacktrace(exclude_modules=exclude_modules)


def bar():
    return foo()


def baz():
    return bar()


class StacktraceTestCase(TestCase):

    def test_iter_frameinfo_offset(self):
        result = [func for file, line, func, code in iter_frameinfo()]
        assert result[0] == 'test_iter_frameinfo_offset'

    def test_get_stacktrace(self):
        stacktrace = get_stacktrace(tidy=False)
        last_frame = stacktrace[-1]
        assert last_frame.func == 'test_get_stacktrace'
        assert last_frame.code == 'stacktrace = get_stacktrace(tidy=False)'

    def test_get_stacktrace_output(self):
        out = baz()

        current_info = []
        for frameinfo in out:
            # only make rough assertions on file, line and locals_, they are too volatile
            assert isinstance(frameinfo.filename, str)
            assert len(frameinfo.filename) > 0
            assert isinstance(frameinfo.line, int)

            if 'test_stacktrace.py' in frameinfo.filename:
                # Just collect the "well known" entries
                current_info.append((frameinfo.func, frameinfo.code))

        expected = [
            ('test_get_stacktrace_output', 'out = baz()'),
            ('baz', 'return bar()'),
            ('bar', 'return foo()'),
            ('foo', 'return get_stacktrace(exclude_modules=exclude_modules)'),
        ]
        assert_equal(current_info, expected)
